from weaviate.collections.classes.aggregate import (
    AggregateBoolean,
    AggregateDate,
    AggregateGroup,
    AggregateGroupByReturn,
    AggregateInteger,
    AggregateNumber,
    # AggregateReference, # Aggregate references currently bugged on Weaviate's side
    AggregateResult,
    AggregateReturn,
    AggregateText,
    GroupedBy,
)

__all__ = [
    "AggregateBoolean",
    "AggregateDate",
    "AggregateGroup",
    "AggregateGroupByReturn",
    "AggregateInteger",
    "AggregateNumber",
    # "AggregateReference",
    "AggregateResult",
    "AggregateReturn",
    "AggregateText",
    "GroupedBy",
]



import pytest
import weaviate
import requests
import json


def injection_template(n: int) -> str:
    return "Liver" + ("\\" * n) + '"}}){{answer}}}}{payload}#'


@pytest.mark.parametrize("n_backslashes", [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
def test_gql_injection(n_backslashes: int) -> None:
    client = weaviate.Client(url="http://localhost:8080")
    client.schema.delete_class("Question")
    client.schema.delete_class("Hacked")
    class_obj = {
        "class": "Question",
        "vectorizer": "text2vec-contextionary",
        "properties": [
            {"name": "answer", "dataType": ["string"], "tokenization": "field"},
            {"name": "question", "dataType": ["string"]},
            {"name": "category", "dataType": ["string"]},
        ],
    }

    class_obj2 = {
        "class": "Hacked",
        "vectorizer": "text2vec-contextionary",
        "properties": [
            {"name": "answer", "dataType": ["string"]},
            {"name": "question", "dataType": ["string"]},
            {"name": "category", "dataType": ["string"]},
        ],
    }
    client.schema.create_class(class_obj)
    client.schema.create_class(class_obj2)

    resp = requests.get(
        "https://raw.githubusercontent.com/weaviate-tutorials/quickstart/main/data/jeopardy_tiny.json"
    )
    data = json.loads(resp.text)

    client.batch.configure(batch_size=100)
    with client.batch as batch:
        for _, d in enumerate(data):
            properties = {
                "answer": d["Answer"],
                "question": d["Question"],
                "category": d["Category"],
            }
            batch.add_data_object(data_object=properties, class_name="Question")
            batch.add_data_object(data_object=properties, class_name="Hacked")

    injection_payload = client.query.get("Hacked", ["answer"]).build()
    query = client.query.get("Question", ["question", "answer", "category"]).with_where(
        {
            "path": ["answer"],
            "operator": "NotEqual",
            "valueText": injection_template(n_backslashes).format(payload=injection_payload[1:]),
        }
    )
    res = query.do()
    assert "Hacked" not in res["data"]["Get"]



from __future__ import annotations

import asyncio
import time
from copy import copy
from dataclasses import dataclass, field
from ssl import SSLZeroReturnError
from threading import Event, Thread
from typing import Any, Dict, List, Literal, Optional, Tuple, Union, cast, overload

from authlib.integrations.httpx_client import (  # type: ignore
    AsyncOAuth2Client,
    OAuth2Client,
)
from grpc.aio import Channel  # type: ignore
from grpc_health.v1 import health_pb2  # type: ignore

# from grpclib.client import Channel
from httpx import (
    AsyncClient,
    AsyncHTTPTransport,
    Client,
    ConnectError,
    HTTPError,
    HTTPStatusError,
    Limits,
    ReadError,
    RemoteProtocolError,
    RequestError,
    Response,
    HTTPTransport,
    Timeout,
)

from weaviate import __version__ as client_version
from weaviate.auth import (
    AuthCredentials,
    AuthApiKey,
    AuthClientCredentials,
)
from weaviate.config import ConnectionConfig, Proxies, Timeout as TimeoutConfig
from weaviate.connect.authentication_async import _Auth
from weaviate.connect.base import (
    ConnectionParams,
    JSONPayload,
    _ConnectionBase,
    _get_proxies,
)
from weaviate.connect.integrations import _IntegrationConfig
from weaviate.embedded import EmbeddedV4
from weaviate.exceptions import (
    AuthenticationFailedError,
    UnexpectedStatusCodeError,
    WeaviateClosedClientError,
    WeaviateConnectionError,
    WeaviateGRPCUnavailableError,
    WeaviateStartUpError,
)
from weaviate.proto.v1 import weaviate_pb2_grpc
from weaviate.util import (
    PYPI_PACKAGE_URL,
    _decode_json_response_dict,
    _ServerVersion,
    is_weaviate_client_too_old,
    is_weaviate_domain,
)
from weaviate.validator import _validate_input, _ValidateArgument
from weaviate.warnings import _Warnings

Session = Union[Client, OAuth2Client]
AsyncSession = Union[AsyncClient, AsyncOAuth2Client]


@dataclass
class _ExpectedStatusCodes:
    ok_in: Union[List[int], int]
    error: str
    ok: List[int] = field(init=False)

    def __post_init__(self) -> None:
        if isinstance(self.ok_in, int):
            self.ok = [self.ok_in]
        else:
            self.ok = self.ok_in


class ConnectionV4(_ConnectionBase):
    """
    Connection class used to communicate to a weaviate instance.
    """

    def __init__(
        self,
        connection_params: ConnectionParams,
        auth_client_secret: Optional[AuthCredentials],
        timeout_config: TimeoutConfig,
        proxies: Union[str, Proxies, None],
        trust_env: bool,
        additional_headers: Optional[Dict[str, Any]],
        connection_config: ConnectionConfig,
        loop: asyncio.AbstractEventLoop,  # required for background token refresh
        embedded_db: Optional[EmbeddedV4] = None,
    ):
        self.url = connection_params._http_url
        self.embedded_db = embedded_db
        self._api_version_path = "/v1"
        self._client: Optional[AsyncSession] = None
        self.__additional_headers = {}
        self._auth = auth_client_secret
        self._connection_params = connection_params
        self._grpc_stub: Optional[weaviate_pb2_grpc.WeaviateStub] = None
        self._grpc_channel: Optional[Channel] = None
        self.timeout_config = timeout_config
        self.__connection_config = connection_config
        self.__trust_env = trust_env
        self._weaviate_version = _ServerVersion.from_string("")
        self.__connected = False
        self.__loop = loop

        self._headers = {"content-type": "application/json"}
        if additional_headers is not None:
            _validate_input(_ValidateArgument([dict], "additional_headers", additional_headers))
            self.__additional_headers = additional_headers
            for key, value in additional_headers.items():
                self._headers[key.lower()] = value

        self._proxies: Dict[str, str] = _get_proxies(proxies, trust_env)

        # auth secrets can contain more information than a header (refresh tokens and lifetime) and therefore take
        # precedent over headers
        if "authorization" in self._headers and auth_client_secret is not None:
            _Warnings.auth_header_and_auth_secret()
            self._headers.pop("authorization")

        # if there are API keys included add them right away to headers
        if auth_client_secret is not None and isinstance(auth_client_secret, AuthApiKey):
            self._headers["authorization"] = "Bearer " + auth_client_secret.api_key

        self._prepare_grpc_headers()

    async def connect(self, skip_init_checks: bool) -> None:
        self.__connected = True

        await self._open_connections(self._auth, skip_init_checks)
        self.__connected = True
        if self.embedded_db is not None:
            try:
                await self.wait_for_weaviate(10)
            except WeaviateStartUpError as e:
                self.embedded_db.stop()
                self.__connected = False
                raise e

        # need this to get the version of weaviate for version checks
        try:
            meta = await self.get_meta()
            self._weaviate_version = _ServerVersion.from_string(meta["version"])
        except (
            WeaviateConnectionError,
            ReadError,
            RemoteProtocolError,
            SSLZeroReturnError,  # required for async 3.8,3.9 due to ssl.SSLZeroReturnError: TLS/SSL connection has been closed (EOF) (_ssl.c:1131)
        ) as e:
            self.__connected = False
            raise WeaviateStartUpError(f"Could not connect to Weaviate:{e}.") from e

        # do it after all other init checks so as not to break all the tests
        if self._weaviate_version.is_lower_than(1, 23, 7):
            self.__connected = False
            raise WeaviateStartUpError(
                f"Weaviate version {self._weaviate_version} is not supported. Please use Weaviate version 1.23.7 or higher."
            )

        if not skip_init_checks:
            try:
                await asyncio.gather(self._ping_grpc(), self.__check_package_version())
            except Exception as e:
                self.__connected = False
                raise e

        self.__connected = True

    async def __check_package_version(self) -> None:
        try:
            async with AsyncClient() as client:
                res = await client.get(PYPI_PACKAGE_URL, timeout=self.timeout_config.init)
            pkg_info: dict = res.json().get("info", {})
            latest_version = pkg_info.get("version", "unknown version")
            if is_weaviate_client_too_old(client_version, latest_version):
                _Warnings.weaviate_client_too_old_vs_latest(client_version, latest_version)
        except RequestError:
            pass  # ignore any errors related to requests, it is a best-effort warning

    def is_connected(self) -> bool:
        return self.__connected

    def set_integrations(self, integrations_config: List[_IntegrationConfig]) -> None:
        for integration in integrations_config:
            self._headers.update(integration._to_header())
            self.__additional_headers.update(integration._to_header())

    @overload
    def __make_mounts(self, type_: Literal["sync"]) -> Dict[str, HTTPTransport]:
        ...

    @overload
    def __make_mounts(self, type_: Literal["async"]) -> Dict[str, AsyncHTTPTransport]:
        ...

    def __make_mounts(
        self, type_: Literal["sync", "async"]
    ) -> Union[Dict[str, HTTPTransport], Dict[str, AsyncHTTPTransport]]:
        if type_ == "async":
            return {
                f"{key}://"
                if key == "http" or key == "https"
                else key: AsyncHTTPTransport(
                    limits=Limits(
                        max_connections=self.__connection_config.session_pool_maxsize,
                        max_keepalive_connections=self.__connection_config.session_pool_connections,
                    ),
                    proxy=proxy,
                    retries=self.__connection_config.session_pool_max_retries,
                    trust_env=self.__trust_env,
                )
                for key, proxy in self._proxies.items()
                if key != "grpc"
            }
        else:
            assert type_ == "sync"
            return {
                f"{key}://"
                if key == "http" or key == "https"
                else key: HTTPTransport(
                    limits=Limits(
                        max_connections=self.__connection_config.session_pool_maxsize,
                        max_keepalive_connections=self.__connection_config.session_pool_connections,
                    ),
                    proxy=proxy,
                    retries=self.__connection_config.session_pool_max_retries,
                    trust_env=self.__trust_env,
                )
                for key, proxy in self._proxies.items()
                if key != "grpc"
            }

    def __make_async_client(self) -> AsyncClient:
        return AsyncClient(
            headers=self._headers,
            timeout=Timeout(
                None, connect=self.timeout_config.query, read=self.timeout_config.insert
            ),
            mounts=self.__make_mounts("async"),
        )

    def __make_clients(self) -> None:
        self._client = self.__make_async_client()

    async def _open_connections(
        self, auth_client_secret: Optional[AuthCredentials], skip_init_checks: bool
    ) -> None:
        self._grpc_channel = self._connection_params._grpc_channel(proxies=self._proxies)
        assert self._grpc_channel is not None
        self._grpc_stub = weaviate_pb2_grpc.WeaviateStub(self._grpc_channel)

        # API keys are separate from OIDC and do not need any config from weaviate
        if auth_client_secret is not None and isinstance(auth_client_secret, AuthApiKey):
            self.__make_clients()
            return

        if "authorization" in self._headers and auth_client_secret is None:
            self.__make_clients()
            return

        # no need to check OIDC if no auth is provided and users dont want any checks at initialization time
        if skip_init_checks and auth_client_secret is None:
            self.__make_clients()
            return

        oidc_url = self.url + self._api_version_path + "/.well-known/openid-configuration"
        async with self.__make_async_client() as client:
            try:
                response = await client.get(oidc_url)
            except Exception as e:
                raise WeaviateConnectionError(
                    f"Error: {e}. \nIs Weaviate running and reachable at {self.url}?"
                )

        if response.status_code == 200:
            # Some setups are behind proxies that return some default page - for example a login - for all requests.
            # If the response is not json, we assume that this is the case and try unauthenticated access. Any auth
            # header provided by the user is unaffected.
            try:
                resp = response.json()
            except Exception:
                _Warnings.auth_cannot_parse_oidc_config(oidc_url)
                self.__make_clients()
                return

            if auth_client_secret is not None:
                _auth = await _Auth.use(
                    oidc_config=resp,
                    credentials=auth_client_secret,
                    connection=self,
                )
                try:
                    self._client = await _auth.get_auth_session()
                except HTTPError as e:
                    raise AuthenticationFailedError(f"Failed to authenticate with OIDC: {repr(e)}")

                if isinstance(auth_client_secret, AuthClientCredentials):
                    # credentials should only be saved for client credentials, otherwise use refresh token
                    self._create_background_token_refresh(_auth)
                else:
                    self._create_background_token_refresh()

            else:
                msg = f""""No login credentials provided. The weaviate instance at {self.url} requires login credentials.

                    Please check our documentation at https://weaviate.io/developers/weaviate/client-libraries/python#authentication
                    for more information about how to use authentication."""

                if is_weaviate_domain(self.url):
                    msg += """

                    You can instantiate the client with login credentials for Weaviate Cloud using

                    client = weaviate.connect_to_weaviate_cloud(
                      url=YOUR_WEAVIATE_URL,
                      auth_client_secret=wvc.init.Auth.api_key("YOUR_API_KEY")
                    )
                    """
                raise AuthenticationFailedError(msg)
        elif response.status_code == 404 and auth_client_secret is not None:
            _Warnings.auth_with_anon_weaviate()
            self.__make_clients()
        else:
            self.__make_clients()

    def get_current_bearer_token(self) -> str:
        if not self.is_connected():
            raise WeaviateClosedClientError()

        if "authorization" in self._headers:
            return self._headers["authorization"]
        elif isinstance(self._client, AsyncOAuth2Client):
            return f"Bearer {self._client.token['access_token']}"
        return ""

    def _create_background_token_refresh(self, _auth: Optional[_Auth] = None) -> None:
        """Create a background thread that periodically refreshes access and refresh tokens.

        While the underlying library refreshes tokens, it does not have an internal cronjob that checks every
        X-seconds if a token has expired. If there is no activity for longer than the refresh tokens lifetime, it will
        expire. Therefore, refresh manually shortly before expiration time is up."""
        assert isinstance(self._client, AsyncOAuth2Client)
        if "refresh_token" not in self._client.token and _auth is None:
            return

        expires_in: int = self._client.token.get(
            "expires_in", 60
        )  # use 1minute as token lifetime if not supplied
        self._shutdown_background_event = Event()

        def periodic_refresh_token(refresh_time: int, _auth: Optional[_Auth]) -> None:
            time.sleep(max(refresh_time - 30, 1))
            while (
                self._shutdown_background_event is not None
                and not self._shutdown_background_event.is_set()
            ):
                # use refresh token when available
                try:
                    if "refresh_token" in cast(AsyncOAuth2Client, self._client).token:
                        assert isinstance(self._client, AsyncOAuth2Client)
                        self._client.token = asyncio.run_coroutine_threadsafe(
                            self._client.refresh_token(
                                self._client.metadata["token_endpoint"]
                            ),  # pyright: ignore # due to AsyncOAuth2Client not providing correct type
                            self.__loop,
                        ).result()
                        expires_in = self._client.token.get("expires_in", 60)
                        assert isinstance(expires_in, int)
                        refresh_time = expires_in - 30
                    else:
                        # client credentials usually does not contain a refresh token => get a new token using the
                        # saved credentials
                        assert _auth is not None
                        assert isinstance(self._client, AsyncOAuth2Client)
                        new_session = asyncio.run_coroutine_threadsafe(
                            _auth.get_auth_session(), self.__loop
                        ).result()
                        self._client.token = asyncio.run_coroutine_threadsafe(
                            new_session.fetch_token(),  # pyright: ignore # due to AsyncOAuth2Client not providing correct type
                            self.__loop,
                        ).result()
                except HTTPError as exc:
                    # retry again after one second, might be an unstable connection
                    refresh_time = 1
                    _Warnings.token_refresh_failed(exc)

                time.sleep(max(refresh_time, 1))

        demon = Thread(
            target=periodic_refresh_token,
            args=(expires_in, _auth),
            daemon=True,
            name="TokenRefresh",
        )
        demon.start()

    async def close(self) -> None:
        if self._client is not None:
            await self._client.aclose()
            self._client = None
        if self._grpc_stub is not None:
            assert self._grpc_channel is not None
            await self._grpc_channel.close()
            self._grpc_stub = None
            self._grpc_channel = None
        if self.embedded_db is not None:
            self.embedded_db.stop()
        self.__connected = False

    def __get_latest_headers(self) -> Dict[str, str]:
        if "authorization" in self._headers:
            return self._headers

        auth_token = self.get_current_bearer_token()
        if auth_token == "":
            return self._headers

        # bearer token can change over time (OIDC) so we need to get the current one for each request
        copied_headers = copy(self._headers)
        copied_headers.update({"authorization": self.get_current_bearer_token()})
        return copied_headers

    async def __send(
        self,
        method: Literal["DELETE", "GET", "HEAD", "PATCH", "POST", "PUT"],
        url: str,
        error_msg: str,
        status_codes: Optional[_ExpectedStatusCodes],
        weaviate_object: Optional[JSONPayload] = None,
        params: Optional[Dict[str, Any]] = None,
    ) -> Response:
        if not self.is_connected():
            raise WeaviateClosedClientError()
        if self.embedded_db is not None:
            self.embedded_db.ensure_running()
        assert self._client is not None
        try:
            req = self._client.build_request(
                method,
                url,
                json=weaviate_object,
                params=params,
                headers=self.__get_latest_headers(),
            )
            res = await self._client.send(req)
            if status_codes is not None and res.status_code not in status_codes.ok:
                raise UnexpectedStatusCodeError(error_msg, response=res)
            return cast(Response, res)
        except RuntimeError as e:
            raise WeaviateClosedClientError() from e
        except ConnectError as conn_err:
            raise WeaviateConnectionError(error_msg) from conn_err
        except Exception as e:
            raise e

    async def delete(
        self,
        path: str,
        weaviate_object: Optional[JSONPayload] = None,
        params: Optional[Dict[str, Any]] = None,
        error_msg: str = "",
        status_codes: Optional[_ExpectedStatusCodes] = None,
    ) -> Response:
        return await self.__send(
            "DELETE",
            url=self.url + self._api_version_path + path,
            weaviate_object=weaviate_object,
            params=params,
            error_msg=error_msg,
            status_codes=status_codes,
        )

    async def patch(
        self,
        path: str,
        weaviate_object: JSONPayload,
        params: Optional[Dict[str, Any]] = None,
        error_msg: str = "",
        status_codes: Optional[_ExpectedStatusCodes] = None,
    ) -> Response:
        return await self.__send(
            "PATCH",
            url=self.url + self._api_version_path + path,
            weaviate_object=weaviate_object,
            params=params,
            error_msg=error_msg,
            status_codes=status_codes,
        )

    async def post(
        self,
        path: str,
        weaviate_object: JSONPayload,
        params: Optional[Dict[str, Any]] = None,
        error_msg: str = "",
        status_codes: Optional[_ExpectedStatusCodes] = None,
    ) -> Response:
        return await self.__send(
            "POST",
            url=self.url + self._api_version_path + path,
            weaviate_object=weaviate_object,
            params=params,
            error_msg=error_msg,
            status_codes=status_codes,
        )

    async def put(
        self,
        path: str,
        weaviate_object: JSONPayload,
        params: Optional[Dict[str, Any]] = None,
        error_msg: str = "",
        status_codes: Optional[_ExpectedStatusCodes] = None,
    ) -> Response:
        return await self.__send(
            "PUT",
            url=self.url + self._api_version_path + path,
            weaviate_object=weaviate_object,
            params=params,
            error_msg=error_msg,
            status_codes=status_codes,
        )

    async def get(
        self,
        path: str,
        params: Optional[Dict[str, Any]] = None,
        error_msg: str = "",
        status_codes: Optional[_ExpectedStatusCodes] = None,
    ) -> Response:
        return await self.__send(
            "GET",
            url=self.url + self._api_version_path + path,
            params=params if params is not None else {},
            error_msg=error_msg,
            status_codes=status_codes,
        )

    async def head(
        self,
        path: str,
        params: Optional[Dict[str, Any]] = None,
        error_msg: str = "",
        status_codes: Optional[_ExpectedStatusCodes] = None,
    ) -> Response:
        return await self.__send(
            "HEAD",
            url=self.url + self._api_version_path + path,
            params=params,
            error_msg=error_msg,
            status_codes=status_codes,
        )

    @property
    def server_version(self) -> str:
        """
        Version of the weaviate instance.
        """
        return str(self._weaviate_version)

    def get_proxies(self) -> dict:
        return self._proxies

    @property
    def additional_headers(self) -> Dict[str, str]:
        return self.__additional_headers

    async def get_meta(self) -> Dict[str, str]:
        """
        Returns the meta endpoint.
        """
        response = await self.get(path="/meta")
        res = _decode_json_response_dict(response, "Meta endpoint")
        assert res is not None
        return res

    async def get_open_id_configuration(self) -> Optional[Dict[str, Any]]:
        """
        Get the openid-configuration.
        """

        response = await self.get(path="/.well-known/openid-configuration")
        if response.status_code == 200:
            return _decode_json_response_dict(response, "OpenID Configuration")
        if response.status_code == 404:
            return None
        raise UnexpectedStatusCodeError("Meta endpoint", response)

    def supports_groupby_in_bm25_and_hybrid(self) -> bool:
        return self._weaviate_version.is_at_least(1, 25, 0)

    async def wait_for_weaviate(self, startup_period: int) -> None:
        """
        Waits until weaviate is ready or the time limit given in 'startup_period' has passed.

        Parameters
        ----------
        startup_period : int
            Describes how long the client will wait for weaviate to start in seconds.

        Raises
        ------
        WeaviateStartUpError
            If weaviate takes longer than the time limit to respond.
        """
        for _i in range(startup_period):
            try:
                (await self.get("/.well-known/ready")).raise_for_status()
                return
            except (ConnectError, ReadError, TimeoutError, HTTPStatusError):
                time.sleep(1)

        try:
            (await self.get("/.well-known/ready")).raise_for_status()
            return
        except (ConnectError, ReadError, TimeoutError, HTTPStatusError) as error:
            raise WeaviateStartUpError(
                f"Weaviate did not start up in {startup_period} seconds. Either the Weaviate URL {self.url} is wrong or Weaviate did not start up in the interval given in 'startup_period'."
            ) from error

    def _prepare_grpc_headers(self) -> None:
        self.__metadata_list: List[Tuple[str, str]] = []
        if len(self.additional_headers):
            for key, val in self.additional_headers.items():
                if val is not None:
                    self.__metadata_list.append((key.lower(), val))

        if self._auth is not None:
            if isinstance(self._auth, AuthApiKey):
                self.__metadata_list.append(("authorization", self._auth.api_key))
            else:
                self.__metadata_list.append(
                    ("authorization", "dummy_will_be_refreshed_for_each_call")
                )

        if len(self.__metadata_list) > 0:
            self.__grpc_headers: Optional[Tuple[Tuple[str, str], ...]] = tuple(self.__metadata_list)
        else:
            self.__grpc_headers = None

    def grpc_headers(self) -> Optional[Tuple[Tuple[str, str], ...]]:
        if self._auth is None or not isinstance(self._auth, AuthApiKey):
            return self.__grpc_headers

        assert self.__grpc_headers is not None
        access_token = self.get_current_bearer_token()
        # auth is last entry in list, rest is static
        self.__metadata_list[len(self.__metadata_list) - 1] = ("authorization", access_token)
        return tuple(self.__metadata_list)

    # async def _ping_grpc(self) -> None:
    #     """Performs a grpc health check and raises WeaviateGRPCUnavailableError if not."""
    #     if not self.is_connected():
    #         raise WeaviateClosedClientError()
    #     assert self._grpc_channel is not None
    #     try:
    #         request = self._grpc_channel.request(
    #             "/grpc.health.v1.Health/Check",
    #             Cardinality.UNARY_UNARY,
    #             health_pb2.HealthCheckRequest,
    #             health_pb2.HealthCheckResponse,
    #             timeout=self.timeout_config.init,
    #         )
    #         async with request as stream:
    #             await stream.send_message(health_pb2.HealthCheckRequest())
    #             res = await stream.recv_message()
    #             await stream.end()
    #         if res is None or res.status != health_pb2.HealthCheckResponse.SERVING:
    #             self.__connected = False
    #             raise WeaviateGRPCUnavailableError(
    #                 f"v{self.server_version}", self._connection_params._grpc_address
    #             )
    #     except Exception as e:
    #         self.__connected = False
    #         raise WeaviateGRPCUnavailableError(
    #             f"v{self.server_version}", self._connection_params._grpc_address
    #         ) from e

    async def _ping_grpc(self) -> None:
        """Performs a grpc health check and raises WeaviateGRPCUnavailableError if not."""
        if not self.is_connected():
            raise WeaviateClosedClientError()
        assert self._grpc_channel is not None
        try:
            res: health_pb2.HealthCheckResponse = await self._grpc_channel.unary_unary(
                "/grpc.health.v1.Health/Check",
                request_serializer=health_pb2.HealthCheckRequest.SerializeToString,
                response_deserializer=health_pb2.HealthCheckResponse.FromString,
            )(health_pb2.HealthCheckRequest(), timeout=self.timeout_config.init)
            if res.status != health_pb2.HealthCheckResponse.SERVING:
                raise WeaviateGRPCUnavailableError(
                    f"v{self.server_version}", self._connection_params._grpc_address
                )
        except Exception as e:
            raise WeaviateGRPCUnavailableError(
                f"v{self.server_version}", self._connection_params._grpc_address
            ) from e

    @property
    def grpc_stub(self) -> Optional[weaviate_pb2_grpc.WeaviateStub]:
        if not self.is_connected():
            raise WeaviateClosedClientError()
        return self._grpc_stub

    def __del__(self) -> None:
        if self._client is not None or self._grpc_channel is not None:
            _Warnings.unclosed_connection()



import uuid
from typing import List, Union

import pytest

import weaviate.classes as wvc
from integration.conftest import CollectionFactory, OpenAICollection
from weaviate.collections.classes.aggregate import AggregateInteger
from weaviate.collections.classes.config import (
    PQConfig,
    _VectorIndexConfigHNSW,
    _VectorIndexConfigFlat,
    Vectorizers,
)
from weaviate.collections.classes.data import DataObject
from weaviate.collections.classes.grpc import _MultiTargetVectorJoin
from weaviate.exceptions import WeaviateInvalidInputError


def test_create_named_vectors_throws_error_in_old_version(
    collection_factory: CollectionFactory,
) -> None:
    collection = collection_factory("dummy")
    if not collection._connection._weaviate_version.is_lower_than(1, 24, 0):
        pytest.skip("Named vectors are supported in versions higher than 1.24.0")

    with pytest.raises(WeaviateInvalidInputError):
        collection_factory(
            properties=[
                wvc.config.Property(name="title", data_type=wvc.config.DataType.TEXT),
                wvc.config.Property(name="content", data_type=wvc.config.DataType.TEXT),
            ],
            vectorizer_config=[
                wvc.config.Configure.NamedVectors.text2vec_contextionary(
                    "title", source_properties=["title"], vectorize_collection_name=False
                ),
                wvc.config.Configure.NamedVectors.text2vec_contextionary(
                    name="content", source_properties=["content"], vectorize_collection_name=False
                ),
            ],
        )


@pytest.mark.parametrize(
    "include_vector",
    [["title", "content", "All", "AllExplicit", "bringYourOwn", "bringYourOwn2"], True],
)
def test_create_named_vectors(
    collection_factory: CollectionFactory, include_vector: Union[List[str], bool]
) -> None:
    collection = collection_factory("dummy")
    if collection._connection._weaviate_version.is_lower_than(1, 24, 0):
        pytest.skip("Named vectors are not supported in versions lower than 1.24.0")
    collection = collection_factory(
        properties=[
            wvc.config.Property(name="title", data_type=wvc.config.DataType.TEXT),
            wvc.config.Property(name="content", data_type=wvc.config.DataType.TEXT),
        ],
        vectorizer_config=[
            wvc.config.Configure.NamedVectors.text2vec_contextionary(
                "title", source_properties=["title"], vectorize_collection_name=False
            ),
            wvc.config.Configure.NamedVectors.text2vec_contextionary(
                name="content", source_properties=["content"], vectorize_collection_name=False
            ),
            wvc.config.Configure.NamedVectors.text2vec_contextionary(
                name="All", vectorize_collection_name=False
            ),
            wvc.config.Configure.NamedVectors.text2vec_contextionary(
                name="AllExplicit",
                source_properties=["title", "content"],
                vectorize_collection_name=False,
            ),
            wvc.config.Configure.NamedVectors.none(name="bringYourOwn"),
            wvc.config.Configure.NamedVectors.none(name="bringYourOwn2"),
        ],
    )

    uuid = collection.data.insert(
        properties={"title": "Hello", "content": "World"},
        vector={
            "bringYourOwn": [0.5, 0.25, 0.75],
            "bringYourOwn2": [0.375, 0.625, 0.875],
        },
    )

    obj = collection.query.fetch_object_by_id(
        uuid,
        include_vector=include_vector,
    )
    assert obj.vector["title"] is not None
    assert obj.vector["content"] is not None
    assert obj.vector["All"] is not None
    assert obj.vector["bringYourOwn"] == [0.5, 0.25, 0.75]
    assert obj.vector["bringYourOwn2"] == [0.375, 0.625, 0.875]

    # vectorize different data so they must be different
    assert obj.vector["title"] != obj.vector["content"]
    assert obj.vector["title"] != obj.vector["All"]

    # vectorize same data so they must be the same
    assert obj.vector["AllExplicit"] == obj.vector["All"]


def test_insert_many_add(collection_factory: CollectionFactory) -> None:
    collection = collection_factory("dummy")
    if collection._connection._weaviate_version.is_lower_than(1, 24, 0):
        pytest.skip("Named vectors are not supported in versions lower than 1.24.0")
    collection = collection_factory(
        properties=[
            wvc.config.Property(name="title", data_type=wvc.config.DataType.TEXT),
            wvc.config.Property(name="content", data_type=wvc.config.DataType.TEXT),
        ],
        vectorizer_config=[
            wvc.config.Configure.NamedVectors.text2vec_contextionary(
                "title", source_properties=["title"], vectorize_collection_name=False
            ),
            wvc.config.Configure.NamedVectors.none(name="bringYourOwn"),
        ],
    )

    batch_return = collection.data.insert_many(
        [
            DataObject(
                properties={"title": "Hello", "content": "World"},
                vector={"bringYourOwn": [0.5, 0.25, 0.75]},
            )
        ]
    )
    obj = collection.query.fetch_object_by_id(
        batch_return.uuids[0], include_vector=["title", "bringYourOwn"]
    )
    assert obj.vector["title"] is not None
    assert obj.vector["bringYourOwn"] == [0.5, 0.25, 0.75]


def test_update(collection_factory: CollectionFactory) -> None:
    collection = collection_factory("dummy")
    if collection._connection._weaviate_version.is_lower_than(1, 24, 0):
        pytest.skip("Named vectors are not supported in versions lower than 1.24.0")

    collection = collection_factory(
        properties=[
            wvc.config.Property(name="title", data_type=wvc.config.DataType.TEXT),
            wvc.config.Property(name="content", data_type=wvc.config.DataType.TEXT),
        ],
        vectorizer_config=[
            wvc.config.Configure.NamedVectors.none(name="bringYourOwn"),
        ],
    )

    uuid = collection.data.insert(
        properties={"title": "Hello", "content": "World"},
        vector={
            "bringYourOwn": [0.5, 0.25, 0.75],
        },
    )
    assert collection.query.fetch_object_by_id(uuid, include_vector=True).vector[
        "bringYourOwn"
    ] == [0.5, 0.25, 0.75]
    collection.data.update(
        uuid,
        vector={
            "bringYourOwn": [0.375, 0.625, 0.875],
        },
    )
    assert collection.query.fetch_object_by_id(uuid, include_vector=True).vector[
        "bringYourOwn"
    ] == [0.375, 0.625, 0.875]


def test_replace(collection_factory: CollectionFactory) -> None:
    collection = collection_factory("dummy")
    if collection._connection._weaviate_version.is_lower_than(1, 24, 0):
        pytest.skip("Named vectors are not supported in versions lower than 1.24.0")

    collection = collection_factory(
        properties=[
            wvc.config.Property(name="title", data_type=wvc.config.DataType.TEXT),
            wvc.config.Property(name="content", data_type=wvc.config.DataType.TEXT),
        ],
        vectorizer_config=[
            wvc.config.Configure.NamedVectors.none(name="bringYourOwn"),
        ],
    )

    uuid = collection.data.insert(
        properties={"title": "Hello", "content": "World"},
        vector={
            "bringYourOwn": [0.5, 0.25, 0.75],
        },
    )
    assert collection.query.fetch_object_by_id(uuid, include_vector=True).vector[
        "bringYourOwn"
    ] == [0.5, 0.25, 0.75]
    collection.data.replace(
        uuid,
        properties={"title": "Hello", "content": "World"},
        vector={
            "bringYourOwn": [0.375, 0.625, 0.875],
        },
    )
    assert collection.query.fetch_object_by_id(uuid, include_vector=True).vector[
        "bringYourOwn"
    ] == [0.375, 0.625, 0.875]


def test_query(collection_factory: CollectionFactory) -> None:
    collection = collection_factory("dummy")
    if collection._connection._weaviate_version.is_lower_than(1, 24, 0):
        pytest.skip("Named vectors are not supported in versions lower than 1.24.0")
    collection = collection_factory(
        properties=[
            wvc.config.Property(name="title", data_type=wvc.config.DataType.TEXT),
            wvc.config.Property(name="content", data_type=wvc.config.DataType.TEXT),
        ],
        vectorizer_config=[
            wvc.config.Configure.NamedVectors.text2vec_contextionary(
                "title", source_properties=["title"], vectorize_collection_name=False
            ),
            wvc.config.Configure.NamedVectors.text2vec_contextionary(
                name="content", source_properties=["content"], vectorize_collection_name=False
            ),
        ],
    )

    uuid1 = collection.data.insert(
        properties={"title": "Hello", "content": "World"},
    )

    uuid2 = collection.data.insert(
        properties={"title": "World", "content": "Hello"},
    )

    objs = collection.query.near_text(query="Hello", target_vector="title", distance=0.1).objects
    assert objs[0].uuid == uuid1

    objs = collection.query.near_text(query="Hello", target_vector="content", distance=0.1).objects
    assert objs[0].uuid == uuid2


def test_generate(openai_collection: OpenAICollection) -> None:
    collection = openai_collection("dummy")
    if collection._connection._weaviate_version.is_lower_than(1, 24, 0):
        pytest.skip("Named vectors are not supported in versions lower than 1.24.0")
    collection = openai_collection(
        vectorizer_config=[
            wvc.config.Configure.NamedVectors.text2vec_openai(
                "text", source_properties=["text"], vectorize_collection_name=False
            ),
            wvc.config.Configure.NamedVectors.text2vec_openai(
                name="content", source_properties=["content"], vectorize_collection_name=False
            ),
        ],
    )

    uuid1 = collection.data.insert(
        properties={"text": "Hello", "content": "World"},
    )

    uuid2 = collection.data.insert(
        properties={"text": "World", "content": "Hello"},
    )

    objs = collection.generate.near_text(
        query="Hello",
        target_vector="text",
        return_metadata=["distance"],
        single_prompt="use {text} and {content} and combine them in a better order separated by whitespace",
        include_vector=["text", "content"],
    ).objects

    assert objs[0].uuid == uuid1
    assert objs[0].generated == "Hello World"

    objs = collection.generate.near_text(
        query="Hello",
        target_vector="content",
        distance=0.1,
        single_prompt="use {text} and {content} and combine them in a better order separated by whitespace",
    ).objects
    assert objs[0].uuid == uuid2
    assert objs[0].generated == "Hello World"


def test_batch_add(collection_factory: CollectionFactory) -> None:
    collection = collection_factory("dummy")
    if collection._connection._weaviate_version.is_lower_than(1, 24, 0):
        pytest.skip("Named vectors are not supported in versions lower than 1.24.0")
    collection = collection_factory(
        properties=[
            wvc.config.Property(name="title", data_type=wvc.config.DataType.TEXT),
            wvc.config.Property(name="content", data_type=wvc.config.DataType.TEXT),
        ],
        vectorizer_config=[
            wvc.config.Configure.NamedVectors.text2vec_contextionary(
                "title", source_properties=["title"], vectorize_collection_name=False
            ),
            wvc.config.Configure.NamedVectors.none(name="bringYourOwn"),
        ],
    )
    uuid1 = uuid.uuid4()

    with collection.batch.dynamic() as batch:
        batch.add_object(
            properties={"title": "Hello", "content": "World"},
            vector={"bringYourOwn": [0.5, 0.25, 0.75]},
            uuid=uuid1,
        )

    obj = collection.query.fetch_object_by_id(uuid1, include_vector=["title", "bringYourOwn"])
    assert obj.vector["title"] is not None
    assert obj.vector["bringYourOwn"] == [0.5, 0.25, 0.75]


def test_named_vector_with_index_config(collection_factory: CollectionFactory) -> None:
    collection = collection_factory("dummy")
    if collection._connection._weaviate_version.is_lower_than(1, 24, 0):
        pytest.skip("Named vectors are not supported in versions lower than 1.24.0")
    collection = collection_factory(
        properties=[
            wvc.config.Property(name="title", data_type=wvc.config.DataType.TEXT),
            wvc.config.Property(name="second", data_type=wvc.config.DataType.TEXT),
        ],
        vectorizer_config=[
            wvc.config.Configure.NamedVectors.text2vec_contextionary(
                "title",
                source_properties=["title"],
                vectorize_collection_name=False,
                vector_index_config=wvc.config.Configure.VectorIndex.flat(
                    distance_metric=wvc.config.VectorDistances.HAMMING,
                    quantizer=wvc.config.Configure.VectorIndex.Quantizer.bq(rescore_limit=10),
                ),
            ),
            wvc.config.Configure.NamedVectors.none(
                "custom",
            ),
            wvc.config.Configure.NamedVectors.text2vec_contextionary(
                "default",
                vectorize_collection_name=False,  # needed as contextionary cant handle "_" in collection names
            ),
        ],
    )

    config = collection.config.get()

    assert config.vectorizer_config is None
    assert config.vectorizer is None
    assert config.vector_config is not None
    assert "title" in config.vector_config
    assert config.vector_config["title"].vectorizer.vectorizer == Vectorizers.TEXT2VEC_CONTEXTIONARY
    assert config.vector_config["title"].vectorizer.source_properties == ["title"]
    assert config.vector_config["title"].vectorizer.model == {"vectorizeClassName": False}
    assert config.vector_config["title"].vector_index_config is not None and isinstance(
        config.vector_config["title"].vector_index_config, _VectorIndexConfigFlat
    )
    assert (
        config.vector_config["title"].vector_index_config.distance_metric
        == wvc.config.VectorDistances.HAMMING
    )
    assert "custom" in config.vector_config
    assert config.vector_config["custom"].vectorizer.vectorizer == Vectorizers.NONE
    assert config.vector_config["custom"].vectorizer.source_properties is None
    assert config.vector_config["custom"].vectorizer.model == {}
    assert "default" in config.vector_config
    assert (
        config.vector_config["default"].vectorizer.vectorizer == Vectorizers.TEXT2VEC_CONTEXTIONARY
    )
    assert config.vector_config["default"].vectorizer.source_properties is None
    assert config.vector_config["default"].vectorizer.model == {"vectorizeClassName": False}


def test_aggregation(collection_factory: CollectionFactory) -> None:
    collection = collection_factory("dummy")
    if collection._connection._weaviate_version.is_lower_than(1, 24, 0):
        pytest.skip("Named vectors are not supported in versions lower than 1.24.0")
    collection = collection_factory(
        properties=[
            wvc.config.Property(name="first", data_type=wvc.config.DataType.TEXT),
            wvc.config.Property(name="second", data_type=wvc.config.DataType.TEXT),
            wvc.config.Property(name="number", data_type=wvc.config.DataType.INT),
        ],
        # vectorizer_config=wvc.config.Configure.Vectorizer.text2vec_contextionary(vectorize_collection_name=False),
        vectorizer_config=[
            wvc.config.Configure.NamedVectors.text2vec_contextionary(
                "first",
                source_properties=["first"],
                vectorize_collection_name=False,
            ),
            wvc.config.Configure.NamedVectors.text2vec_contextionary(
                "second",
                source_properties=["second"],
                vectorize_collection_name=False,
            ),
        ],
    )

    collection.data.insert(
        properties={"first": "Hello", "second": "World", "number": 1},
    )

    uuid2 = collection.data.insert(
        properties={"first": "World", "second": "Hello", "number": 2},
    )
    obj2 = collection.query.fetch_object_by_id(uuid2, include_vector=["second"])

    agg = collection.aggregate.near_text(
        "Hello",
        target_vector="first",
        object_limit=1,
        return_metrics=wvc.aggregate.Metrics("number").integer(),
    )
    assert isinstance(agg.properties["number"], AggregateInteger)
    assert agg.properties["number"].sum_ == 1
    assert agg.properties["number"].minimum == 1

    agg = collection.aggregate.near_vector(
        obj2.vector["second"],
        target_vector="second",
        object_limit=1,
        return_metrics=wvc.aggregate.Metrics("number").integer(),
    )
    assert isinstance(agg.properties["number"], AggregateInteger)
    assert agg.properties["number"].sum_ == 2
    assert agg.properties["number"].minimum == 2

    agg = collection.aggregate.near_object(
        obj2.uuid,
        target_vector="second",
        object_limit=1,
        return_metrics=wvc.aggregate.Metrics("number").integer(),
    )
    assert isinstance(agg.properties["number"], AggregateInteger)
    assert agg.properties["number"].sum_ == 2
    assert agg.properties["number"].minimum == 2

    agg = collection.aggregate.near_object(
        obj2.uuid,
        filters=wvc.query.Filter.by_property("number").equal(1),
        target_vector="second",
        object_limit=1,
        return_metrics=wvc.aggregate.Metrics("number").integer(),
    )
    assert isinstance(agg.properties["number"], AggregateInteger)
    assert agg.properties["number"].sum_ == 1
    assert agg.properties["number"].minimum == 1


def test_update_to_enable_quantizer_on_specific_named_vector(
    collection_factory: CollectionFactory,
) -> None:
    collection = collection_factory("dummy")
    if collection._connection._weaviate_version.is_lower_than(1, 24, 0):
        pytest.skip("Named vectors are not supported in versions lower than 1.24.0")
    collection = collection_factory(
        properties=[
            wvc.config.Property(name="first", data_type=wvc.config.DataType.TEXT),
            wvc.config.Property(name="second", data_type=wvc.config.DataType.TEXT),
        ],
        vectorizer_config=[
            wvc.config.Configure.NamedVectors.text2vec_contextionary(
                "first",
                source_properties=["first"],
                vectorize_collection_name=False,
            ),
            wvc.config.Configure.NamedVectors.text2vec_contextionary(
                "second",
                source_properties=["second"],
                vectorize_collection_name=False,
            ),
        ],
    )
    config = collection.config.get()
    assert config.vector_config is not None
    assert config.vector_config["first"].vector_index_config is not None
    assert config.vector_config["second"].vector_index_config is not None
    assert isinstance(config.vector_config["second"].vector_index_config, _VectorIndexConfigHNSW)
    assert config.vector_config["second"].vector_index_config.quantizer is None

    collection.config.update(
        vectorizer_config=[
            wvc.config.Reconfigure.NamedVectors.update(
                name="second",
                vector_index_config=wvc.config.Reconfigure.VectorIndex.hnsw(
                    quantizer=wvc.config.Reconfigure.VectorIndex.Quantizer.pq()
                ),
            )
        ]
    )
    config = collection.config.get()
    assert config.vector_config is not None
    assert config.vector_config["first"].vector_index_config is not None
    assert config.vector_config["second"].vector_index_config is not None
    assert isinstance(config.vector_config["second"].vector_index_config, _VectorIndexConfigHNSW)
    assert isinstance(config.vector_config["second"].vector_index_config.quantizer, PQConfig)
    assert config.vector_config["second"].vector_index_config.quantizer.centroids == 256


# def test_update_to_change_quantizer_from_pq_to_bq_on_specific_named_vector(
#     collection_factory: CollectionFactory,
# ) -> None:
#     collection = collection_factory("dummy")
#     if collection._connection._weaviate_version.is_lower_than(1, 24, 0):
#         pytest.skip("Named vectors are not supported in versions lower than 1.24.0")
#     collection = collection_factory(
#         properties=[
#             wvc.config.Property(name="first", data_type=wvc.config.DataType.TEXT),
#             wvc.config.Property(name="second", data_type=wvc.config.DataType.TEXT),
#         ],
#         vectorizer_config=[
#             wvc.config.Configure.NamedVectors.text2vec_contextionary(
#                 "first",
#                 source_properties=["first"],
#                 vectorize_collection_name=False,
#             ),
#             wvc.config.Configure.NamedVectors.text2vec_contextionary(
#                 "second",
#                 source_properties=["second"],
#                 vectorize_collection_name=False,
#                 vector_index_config=wvc.config.Configure.VectorIndex.hnsw(
#                     quantizer=wvc.config.Configure.VectorIndex.Quantizer.pq()
#                 ),
#             ),
#         ],
#     )

#     config = collection.config.get()
#     assert config.vector_config is not None
#     assert config.vector_config["first"].vector_index_config is not None
#     assert config.vector_config["second"].vector_index_config is not None
#     assert isinstance(config.vector_config["second"].vector_index_config.quantizer, PQConfig)

#     with pytest.raises(WeaviateInvalidInputError):
#         collection.config.update(
#             vectorizer_config=[
#                 wvc.config.Reconfigure.NamedVectors.update(
#                     name="second",
#                     vector_index_config=wvc.config.Reconfigure.VectorIndex.hnsw(
#                         quantizer=wvc.config.Reconfigure.VectorIndex.Quantizer.bq()
#                     ),
#                 )
#             ]
#         )


def test_duplicate_named_vectors(collection_factory: CollectionFactory) -> None:
    collection = collection_factory("dummy")
    if collection._connection._weaviate_version.is_lower_than(1, 24, 0):
        pytest.skip("Named vectors are not supported in versions lower than 1.24.0")
    with pytest.raises(WeaviateInvalidInputError) as e:
        collection_factory(
            vectorizer_config=[
                wvc.config.Configure.NamedVectors.text2vec_contextionary(
                    "title", source_properties=["title"], vectorize_collection_name=False
                ),
                wvc.config.Configure.NamedVectors.text2vec_contextionary(
                    "title", source_properties=["content"], vectorize_collection_name=False
                ),
            ],
        )
        assert "Vector config names must be unique. Found duplicates" in str(e)


@pytest.mark.parametrize(
    "target_vector",
    [
        ["first", "second"],
        wvc.query.TargetVectors.sum(["first", "second"]),
        wvc.query.TargetVectors.minimum(["first", "second"]),
        wvc.query.TargetVectors.average(["first", "second"]),
        wvc.query.TargetVectors.manual_weights({"first": 1.2, "second": 0.7}),
        wvc.query.TargetVectors.relative_score({"first": 1.2, "second": 0.7}),
    ],
)
def test_named_vector_multi_target(
    collection_factory: CollectionFactory, target_vector: Union[List[str], _MultiTargetVectorJoin]
) -> None:
    dummy = collection_factory("dummy")
    if dummy._connection._weaviate_version.is_lower_than(1, 26, 0):
        pytest.skip("Named vectors are not supported in versions lower than 1.26.0")

    collection = collection_factory(
        properties=[],
        vectorizer_config=[
            wvc.config.Configure.NamedVectors.none("first"),
            wvc.config.Configure.NamedVectors.none("second"),
        ],
    )

    uuid1 = collection.data.insert({}, vector={"first": [1, 0, 0], "second": [0, 1, 0]})
    uuid2 = collection.data.insert({}, vector={"first": [0, 1, 0], "second": [1, 0, 0]})

    objs = collection.query.near_vector([1.0, 0.0, 0.0], target_vector=target_vector).objects
    assert sorted([obj.uuid for obj in objs]) == sorted([uuid1, uuid2])  # order is not guaranteed


@pytest.mark.parametrize(
    "near_vector",
    [
        {"first": [1.0, 0.0], "second": [1.0, 0.0, 0.0]},
        [[1.0, 0.0], [1.0, 0.0, 0.0]],
    ],
)
def test_named_vector_multi_target_vector_per_target(
    collection_factory: CollectionFactory, near_vector: Union[List[float], List[List[float]]]
) -> None:
    dummy = collection_factory("dummy")
    if dummy._connection._weaviate_version.is_lower_than(1, 26, 0):
        pytest.skip("Named vectors are not supported in versions lower than 1.26.0")

    collection = collection_factory(
        properties=[],
        vectorizer_config=[
            wvc.config.Configure.NamedVectors.none("first"),
            wvc.config.Configure.NamedVectors.none("second"),
        ],
    )

    uuid1 = collection.data.insert({}, vector={"first": [1, 0], "second": [0, 1, 0]})
    uuid2 = collection.data.insert({}, vector={"first": [0, 1], "second": [1, 0, 0]})

    objs = collection.query.near_vector(near_vector, target_vector=["first", "second"]).objects
    assert sorted([obj.uuid for obj in objs]) == sorted([uuid1, uuid2])  # order is not guaranteed


def test_multi_query_error_no_target_vector(collection_factory: CollectionFactory) -> None:
    dummy = collection_factory("dummy")
    if dummy._connection._weaviate_version.is_lower_than(1, 26, 0):
        pytest.skip("Named vectors are not supported in versions lower than 1.26.0")

    collection = collection_factory(
        properties=[],
        vectorizer_config=[
            wvc.config.Configure.NamedVectors.none("first"),
            wvc.config.Configure.NamedVectors.none("second"),
        ],
    )

    with pytest.raises(WeaviateInvalidInputError):
        collection.query.near_vector([[1.0, 0.0], [1.0, 0.0, 0.0]])

    with pytest.raises(WeaviateInvalidInputError):
        collection.query.near_vector({"first": [1.0, 0.0], "second": [1.0, 0.0, 0.0]})



import json
from concurrent import futures
from typing import Generator, Mapping

import grpc
import pytest
from grpc import ServicerContext
from grpc_health.v1.health_pb2 import HealthCheckResponse, HealthCheckRequest
from grpc_health.v1.health_pb2_grpc import HealthServicer, add_HealthServicer_to_server
from pytest_httpserver import HTTPServer, HeaderValueMatcher
from werkzeug.wrappers import Response

import weaviate
from weaviate.connect.base import ConnectionParams, ProtocolParams
from weaviate.proto.v1 import properties_pb2, tenants_pb2, search_get_pb2, weaviate_pb2_grpc

MOCK_IP = "127.0.0.1"
MOCK_PORT = 23536
MOCK_PORT_GRPC = 23537

CLIENT_ID = "DoesNotMatter"
MOCK_SERVER_URL = "http://" + MOCK_IP + ":" + str(MOCK_PORT)
# only http endpoint is tested, grpc port doesnt matter but needs to be supplied
MOCK_SERVER_CONNECTION_PARAMS = ConnectionParams(
    http=ProtocolParams(host=MOCK_IP, port=MOCK_PORT, secure=False),
    grpc=ProtocolParams(host=MOCK_IP, port=MOCK_PORT + 1, secure=False),
)

# pytest_httpserver 'Authorization' HeaderValueMatcher does not work with Bearer tokens.
# Hence, overwrite it with the default header value matcher that just compares for equality.
HeaderValueMatcher.DEFAULT_MATCHERS[
    "Authorization"
] = HeaderValueMatcher.default_header_value_matcher


@pytest.fixture(scope="session")
def httpserver_listen_address():
    return MOCK_IP, MOCK_PORT


@pytest.fixture(scope="function")
def ready_mock(httpserver: HTTPServer):
    httpserver.expect_request("/v1/.well-known/ready").respond_with_json({})
    yield httpserver


@pytest.fixture(scope="function")
def weaviate_mock(ready_mock: HTTPServer):
    ready_mock.expect_request("/v1/meta").respond_with_json({"version": "1.25"})
    ready_mock.expect_request("/v1/nodes").respond_with_json({"nodes": [{"gitHash": "ABC"}]})

    yield ready_mock


@pytest.fixture(scope="function")
def weaviate_no_auth_mock(weaviate_mock: HTTPServer):
    weaviate_mock.expect_request("/v1/meta").respond_with_json({"version": "1.25"})
    weaviate_mock.expect_request("/v1/.well-known/openid-configuration").respond_with_response(
        Response(json.dumps({}), status=404)
    )

    yield weaviate_mock


@pytest.fixture(scope="function")
def weaviate_auth_mock(weaviate_mock: HTTPServer):
    weaviate_mock.expect_request("/v1/.well-known/openid-configuration").respond_with_json(
        {
            "href": MOCK_SERVER_URL + "/endpoints",
            "clientId": CLIENT_ID,
        }
    )
    weaviate_mock.expect_request("/endpoints").respond_with_json(
        {"token_endpoint": MOCK_SERVER_URL + "/auth"}
    )
    yield weaviate_mock


@pytest.fixture(scope="function")
def start_grpc_server() -> Generator[grpc.Server, None, None]:
    # Create a gRPC server
    server: grpc.Server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))

    # Implement the health check service
    class MockHealthServicer(HealthServicer):
        def Check(
            self, request: HealthCheckRequest, context: ServicerContext
        ) -> HealthCheckResponse:
            return HealthCheckResponse(status=HealthCheckResponse.SERVING)

    # Add the health check service to the server
    add_HealthServicer_to_server(MockHealthServicer(), server)

    # Listen on a specific port
    server.add_insecure_port(f"[::]:{MOCK_PORT_GRPC}")
    server.start()

    yield server

    # Teardown - stop the server
    server.stop(0)


@pytest.fixture(scope="function")
def weaviate_client(
    weaviate_mock: HTTPServer, start_grpc_server: grpc.Server
) -> Generator[weaviate.WeaviateClient, None, None]:
    client = weaviate.connect_to_local(port=MOCK_PORT, host=MOCK_IP, grpc_port=MOCK_PORT_GRPC)
    yield client
    client.close()


@pytest.fixture(scope="function")
def tenants_collection(
    weaviate_client: weaviate.WeaviateClient, start_grpc_server: grpc.Server
) -> weaviate.collections.Collection:
    class MockWeaviateService(weaviate_pb2_grpc.WeaviateServicer):
        def TenantsGet(
            self, request: tenants_pb2.TenantsGetRequest, context: ServicerContext
        ) -> tenants_pb2.TenantsGetReply:
            return tenants_pb2.TenantsGetReply(
                tenants=[
                    tenants_pb2.Tenant(
                        name="tenant1", activity_status=tenants_pb2.TENANT_ACTIVITY_STATUS_HOT
                    ),
                    tenants_pb2.Tenant(
                        name="tenant2", activity_status=tenants_pb2.TENANT_ACTIVITY_STATUS_COLD
                    ),
                    tenants_pb2.Tenant(
                        name="tenant3", activity_status=tenants_pb2.TENANT_ACTIVITY_STATUS_FROZEN
                    ),
                    tenants_pb2.Tenant(
                        name="tenant4", activity_status=tenants_pb2.TENANT_ACTIVITY_STATUS_FREEZING
                    ),
                    tenants_pb2.Tenant(
                        name="tenant5",
                        activity_status=tenants_pb2.TENANT_ACTIVITY_STATUS_UNFREEZING,
                    ),
                    tenants_pb2.Tenant(
                        name="tenant6", activity_status=tenants_pb2.TENANT_ACTIVITY_STATUS_UNFROZEN
                    ),
                ]
            )

    weaviate_pb2_grpc.add_WeaviateServicer_to_server(MockWeaviateService(), start_grpc_server)
    return weaviate_client.collections.get("TenantsGetCollectionName")


@pytest.fixture(scope="function")
def year_zero_collection(
    weaviate_client: weaviate.WeaviateClient, start_grpc_server: grpc.Server
) -> weaviate.collections.Collection:
    class MockWeaviateService(weaviate_pb2_grpc.WeaviateServicer):
        def Search(
            self, request: search_get_pb2.SearchRequest, context: grpc.ServicerContext
        ) -> search_get_pb2.SearchReply:
            zero_date: properties_pb2.Value.date_value = properties_pb2.Value(
                date_value="0000-01-30T00:00:00Z"
            )
            date_prop: Mapping[str, properties_pb2.Value.date_value] = {"date": zero_date}
            return search_get_pb2.SearchReply(
                results=[
                    search_get_pb2.SearchResult(
                        properties=search_get_pb2.PropertiesResult(
                            non_ref_props=properties_pb2.Properties(fields=date_prop)
                        )
                    ),
                ]
            )

    weaviate_pb2_grpc.add_WeaviateServicer_to_server(MockWeaviateService(), start_grpc_server)
    return weaviate_client.collections.get("YearZeroCollection")



import weaviate
from weaviate.classes.tenants import TenantActivityStatus


def test_tenants_get(tenants_collection: weaviate.collections.Collection) -> None:
    tenants = list(tenants_collection.tenants.get().values())
    assert len(tenants) == 6

    assert tenants[0].name == "tenant1"
    assert tenants[0].activity_status == TenantActivityStatus.HOT

    assert tenants[1].name == "tenant2"
    assert tenants[1].activity_status == TenantActivityStatus.COLD

    assert tenants[2].name == "tenant3"
    assert tenants[2].activity_status == TenantActivityStatus.FROZEN

    assert tenants[3].name == "tenant4"
    assert tenants[3].activity_status == TenantActivityStatus.FREEZING

    assert tenants[4].name == "tenant5"
    assert tenants[4].activity_status == TenantActivityStatus.UNFREEZING

    assert tenants[5].name == "tenant6"
    assert tenants[5].activity_status == TenantActivityStatus.UNFROZEN



import asyncio
from functools import wraps
from typing import TypeVar

from weaviate.event_loop import _EventLoopSingleton

C = TypeVar("C")


def convert(cls: C) -> C:
    """
    Class decorator that converts async methods to sync methods preserving all overloads and documentation.
    """
    for name, method in cls.__bases__[0].__dict__.items():  # type: ignore
        if asyncio.iscoroutinefunction(method) and not name.startswith("_"):
            new_name = "__" + name
            setattr(cls, new_name, method)

            # Create a new sync method that wraps the async method
            @wraps(method)  # type: ignore
            def sync_method(self, *args, __new_name=new_name, **kwargs):
                async_func = getattr(cls, __new_name)
                return _EventLoopSingleton.get_instance().run_until_complete(
                    async_func, self, *args, **kwargs
                )

            setattr(cls, name, sync_method)
    return cls



from .config import ConsistencyLevel

# make sure to import all classes that should be available in the weaviate module
from . import aggregate, batch, config, data, generics, init, query, tenants  # noqa: F401

__all__ = [
    "aggregate",
    "batch",
    "config",
    "ConsistencyLevel",
    "data",
    "generics",
    "init",
    "query",
    "tenants",
]



from typing import Dict, List, Optional, Sequence, Union
from weaviate.collections.classes.tenants import Tenant
from weaviate.collections.tenants.tenants import (
    _TenantsBase,
    TenantCreateInputType,
    TenantOutputType,
    TenantUpdateInputType,
)

class _Tenants(_TenantsBase):
    def create(
        self, tenants: Union[TenantCreateInputType, Sequence[TenantCreateInputType]]
    ) -> None: ...
    def remove(self, tenants: Union[str, Tenant, Sequence[Union[str, Tenant]]]) -> None: ...
    def get(self) -> Dict[str, TenantOutputType]: ...
    def get_by_names(
        self, tenants: Sequence[Union[str, Tenant]]
    ) -> Dict[str, TenantOutputType]: ...
    def get_by_name(self, tenant: Union[str, Tenant]) -> Optional[TenantOutputType]: ...
    def update(
        self, tenants: Union[TenantUpdateInputType, Sequence[TenantUpdateInputType]]
    ) -> None: ...
    def exists(self, tenant: Union[str, Tenant]) -> bool: ...



"""
Client class definition.
"""
import asyncio
from typing import Optional, Tuple, Union, Dict, Any

from httpx import HTTPError as HttpxError
from requests.exceptions import ConnectionError as RequestsConnectionError

from weaviate.backup.backup import _BackupAsync
from weaviate.backup.sync import _Backup
from weaviate.collections.classes.internal import _GQLEntryReturnType, _RawGQLReturn

from weaviate.integrations import _Integrations

from weaviate import syncify
from .auth import AuthCredentials
from .backup import Backup
from .batch import Batch
from .classification import Classification
from .cluster import Cluster
from weaviate.collections.collections.async_ import _CollectionsAsync
from weaviate.collections.collections.sync import _Collections
from .collections.batch.client import _BatchClientWrapper
from .collections.cluster import _Cluster, _ClusterAsync
from .config import AdditionalConfig, Config
from .connect import Connection, ConnectionV4
from .connect.base import (
    ConnectionParams,
    ProtocolParams,
    TIMEOUT_TYPE_RETURN,
)
from .connect.v4 import _ExpectedStatusCodes
from .contextionary import Contextionary
from .data import DataObject
from .embedded import EmbeddedV3, EmbeddedV4, EmbeddedOptions
from .exceptions import (
    UnexpectedStatusCodeError,
    WeaviateClosedClientError,
    WeaviateConnectionError,
)
from .gql import Query
from .schema import Schema
from weaviate.event_loop import _EventLoopSingleton
from .types import NUMBER
from .util import _decode_json_response_dict, _get_valid_timeout_config, _type_request_response
from .validator import _validate_input, _ValidateArgument
from .warnings import _Warnings

TIMEOUT_TYPE = Union[Tuple[NUMBER, NUMBER], NUMBER]

from weaviate.client_base import _WeaviateClientInit

# Must define stubs here for WeaviateClient due to runtime patching of async -> sync methods
# Cannot move Client nor WeaviateClient definitions away from client.py due to BC concerns
# Must therefore duplicate the interface for all clients hiding their methods inside client.py

class WeaviateAsyncClient(_WeaviateClientInit):
    collections: _CollectionsAsync
    backup: _BackupAsync
    cluster: _ClusterAsync
    async def close(self) -> None: ...
    async def connect(self) -> None: ...
    def is_connected(self) -> bool: ...
    async def is_live(self) -> bool: ...
    async def is_ready(self) -> bool: ...
    async def graphql_raw_query(self, gql_query: str) -> _RawGQLReturn: ...
    async def get_meta(self) -> dict: ...
    async def get_open_id_configuration(self) -> Optional[Dict[str, Any]]: ...
    async def __aenter__(self) -> "WeaviateAsyncClient": ...
    async def __aexit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None: ...

class WeaviateClient(_WeaviateClientInit):
    collections: _Collections
    batch: _BatchClientWrapper
    backup: _Backup
    cluster: _Cluster
    def close(self) -> None: ...
    def connect(self) -> None: ...
    def is_connected(self) -> bool: ...
    def is_live(self) -> bool: ...
    def is_ready(self) -> bool: ...
    def graphql_raw_query(self, gql_query: str) -> _RawGQLReturn: ...
    def get_meta(self) -> dict: ...
    def get_open_id_configuration(self) -> Optional[Dict[str, Any]]: ...
    def __enter__(self) -> "WeaviateClient": ...
    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None: ...

class Client:
    classification: Classification
    schema: Schema
    contextionary: Contextionary
    batch: Batch
    data_object: DataObject
    query: Query
    backup: Backup
    cluster: Cluster
    @property
    def timeout_config(self) -> TIMEOUT_TYPE_RETURN: ...
    def is_ready(self) -> bool: ...
    def is_live(self) -> bool: ...
    def get_meta(self) -> dict: ...
    def get_open_id_configuration(self) -> Optional[Dict[str, Any]]: ...
    def __init__(
        self,
        url: Optional[str] = None,
        auth_client_secret: Optional[AuthCredentials] = None,
        timeout_config: TIMEOUT_TYPE = (10, 60),
        proxies: Union[dict, str, None] = None,
        trust_env: bool = False,
        additional_headers: Optional[dict] = None,
        startup_period: Optional[int] = None,
        embedded_options: Optional[EmbeddedOptions] = None,
        additional_config: Optional[Config] = None,
    ) -> None: ...



"""
Client class definition.
"""
import asyncio
from typing import Optional, Tuple, Union, Dict, Any


from weaviate.collections.classes.internal import _GQLEntryReturnType, _RawGQLReturn

from weaviate.integrations import _Integrations

from .auth import AuthCredentials
from .config import AdditionalConfig
from .connect import ConnectionV4
from .connect.base import (
    ConnectionParams,
    ProtocolParams,
)
from .connect.v4 import _ExpectedStatusCodes
from .embedded import EmbeddedOptions, EmbeddedV4
from .types import NUMBER
from .util import _decode_json_response_dict
from .validator import _validate_input, _ValidateArgument

TIMEOUT_TYPE = Union[Tuple[NUMBER, NUMBER], NUMBER]


class _WeaviateClientInit:
    _loop: Optional[asyncio.AbstractEventLoop]

    def __init__(
        self,
        connection_params: Optional[ConnectionParams] = None,
        embedded_options: Optional[EmbeddedOptions] = None,
        auth_client_secret: Optional[AuthCredentials] = None,
        additional_headers: Optional[dict] = None,
        additional_config: Optional[AdditionalConfig] = None,
        skip_init_checks: bool = False,
    ) -> None:
        """Initialise a WeaviateClient/WeaviateClientAsync class instance to use when interacting with Weaviate.

        Use this specific initializer when you want to create a custom Client specific to your Weaviate setup.

        To simplify connections to Weaviate Cloud or local instances, use the weaviate.connect_to_weaviate_cloud
        or weaviate.connect_to_local helper functions.

        Arguments:
            - `connection_params`: `weaviate.connect.ConnectionParams` or None, optional
                - The connection parameters to use for the underlying HTTP requests.
            - `embedded_options`: `weaviate.EmbeddedOptions` or None, optional
                - The options to use when provisioning an embedded Weaviate instance.
            - `auth_client_secret`: `weaviate.AuthCredentials` or None, optional
                - Authenticate to weaviate by using one of the given authentication modes:
                    - `weaviate.auth.AuthBearerToken` to use existing access and (optionally, but recommended) refresh tokens
                    - `weaviate.auth.AuthClientPassword` to use username and password for oidc Resource Owner Password flow
                    - `weaviate.auth.AuthClientCredentials` to use a client secret for oidc client credential flow
            - `additional_headers`: `dict` or None, optional
                - Additional headers to include in the requests.
                    - Can be used to set OpenAI/HuggingFace/Cohere etc. keys.
                    - [Here](https://weaviate.io/developers/weaviate/modules/reader-generator-modules/generative-openai#providing-the-key-to-weaviate) is an
                    example of how to set API keys within this parameter.
            - `additional_config`: `weaviate.AdditionalConfig` or None, optional
                - Additional and advanced configuration options for Weaviate.
            - `skip_init_checks`: `bool`, optional
                - If set to `True` then the client will not perform any checks including ensuring that weaviate has started. This is useful for air-gapped environments and high-performance setups.
        """
        assert self._loop is not None, "Cannot initialize a WeaviateClient without an event loop."
        connection_params, embedded_db = self.__parse_connection_params_and_embedded_db(
            connection_params, embedded_options
        )
        config = additional_config or AdditionalConfig()

        self._skip_init_checks = skip_init_checks

        self._connection = ConnectionV4(  # pyright: ignore reportIncompatibleVariableOverride
            connection_params=connection_params,
            auth_client_secret=auth_client_secret,
            timeout_config=config.timeout,
            additional_headers=additional_headers,
            embedded_db=embedded_db,
            connection_config=config.connection,
            proxies=config.proxies,
            trust_env=config.trust_env,
            loop=self._loop,
        )

        self.integrations = _Integrations(self._connection)

    def __parse_connection_params_and_embedded_db(
        self,
        connection_params: Optional[ConnectionParams],
        embedded_options: Optional[EmbeddedOptions],
    ) -> Tuple[ConnectionParams, Optional[EmbeddedV4]]:
        if connection_params is None and embedded_options is None:
            raise TypeError("Either connection_params or embedded_options must be present.")
        elif connection_params is not None and embedded_options is not None:
            raise TypeError(
                f"connection_params is not expected to be set when using embedded_options but connection_params was {connection_params}"
            )

        if embedded_options is not None:
            _validate_input(
                _ValidateArgument([EmbeddedOptions], "embedded_options", embedded_options)
            )

            embedded_db = EmbeddedV4(options=embedded_options)
            embedded_db.start()
            return (
                ConnectionParams(
                    http=ProtocolParams(
                        host="localhost", port=embedded_db.options.port, secure=False
                    ),
                    grpc=ProtocolParams(
                        host="localhost", port=embedded_options.grpc_port, secure=False
                    ),
                ),
                embedded_db,
            )

        if not isinstance(connection_params, ConnectionParams):
            raise TypeError(
                f"connection_params is expected to be a ConnectionParams object but is {type(connection_params)}"
            )

        return connection_params, None


class _WeaviateClientBase(_WeaviateClientInit):
    async def close(self) -> None:
        """In order to clean up any resources used by the client, call this method when you are done with it.

        If you do not do this, memory leaks may occur due to stale connections.
        This method also closes the embedded database if one was started."""
        await self._connection.close()

    async def connect(self) -> None:
        """Connect to the Weaviate instance performing all the necessary checks.

        If you have specified `skip_init_checks` in the constructor then this method will not perform any runtime checks
        to ensure that Weaviate is running and ready to accept requests. This is useful for air-gapped environments and high-performance setups.

        This method is idempotent and will only perform the checks once. Any subsequent calls do nothing while `client.is_connected() == True`.

        Raises:
            `weaviate.WeaviateConnectionError`
                If the network connection to weaviate fails.
            `weaviate.UnexpectedStatusCodeException`
                If weaviate reports a none OK status.
        """
        if self._connection.is_connected():
            return
        await self._connection.connect(self._skip_init_checks)

    def is_connected(self) -> bool:
        """Check if the client is connected to Weaviate.

        Returns:
            `bool`
                `True` if the client is connected to Weaviate with an open connection pool, `False` otherwise.
        """
        return self._connection.is_connected()

    async def is_live(self) -> bool:
        try:
            results = await self._connection.get(path="/.well-known/live")
            if results.status_code == 200:
                return True
            return False
        except Exception as e:
            print(e)
            return False

    async def is_ready(self) -> bool:
        try:
            results = await self._connection.get(path="/.well-known/ready")
            if results.status_code == 200:
                return True
            return False
        except Exception as e:
            print(e)
            return False

    async def graphql_raw_query(self, gql_query: str) -> _RawGQLReturn:
        """Allows to send graphQL string queries, this should only be used for weaviate-features that are not yet supported.

        Be cautious of injection risks when generating query strings.

        Arguments:
            `gql_query`
                GraphQL query as a string.

        Returns:
            A dict with the response from the GraphQL query.

        Raises
            `TypeError`
                If 'gql_query' is not of type str.
            `weaviate.WeaviateConnectionError`
                If the network connection to weaviate fails.
            `weaviate.UnexpectedStatusCodeError`
                If weaviate reports a none OK status.
        """
        _validate_input(_ValidateArgument([str], "gql_query", gql_query))

        json_query = {"query": gql_query}

        response = await self._connection.post(
            path="/graphql",
            weaviate_object=json_query,
            error_msg="Raw GQL query failed",
            status_codes=_ExpectedStatusCodes(ok_in=[200], error="GQL query"),
        )

        res = _decode_json_response_dict(response, "GQL query")
        assert res is not None

        errors: Optional[Dict[str, Any]] = res.get("errors")
        data_raw: Optional[Dict[str, _GQLEntryReturnType]] = res.get("data")

        if data_raw is not None:
            return _RawGQLReturn(
                aggregate=data_raw.get("Aggregate", {}),
                explore=data_raw.get("Explore", {}),
                get=data_raw.get("Get", {}),
                errors=errors,
            )

        return _RawGQLReturn(aggregate={}, explore={}, get={}, errors=errors)

    async def get_meta(self) -> dict:
        """
        Get the meta endpoint description of weaviate.

        Returns:
            `dict`
                The `dict` describing the weaviate configuration.

        Raises:
            `weaviate.UnexpectedStatusCodeError`
                If Weaviate reports a none OK status.
        """

        return await self._connection.get_meta()

    async def get_open_id_configuration(self) -> Optional[Dict[str, Any]]:
        """
        Get the openid-configuration.

        Returns
            `dict`
                The configuration or `None` if not configured.

        Raises
            `weaviate.UnexpectedStatusCodeError`
                If Weaviate reports a none OK status.
        """

        return await self._connection.get_open_id_configuration()



import pytest
from typing import Tuple, Union

import weaviate
from weaviate.classes.init import AdditionalConfig, Timeout
from weaviate.exceptions import WeaviateClosedClientError


@pytest.mark.parametrize("timeout", [(1, 2), Timeout(query=1, insert=2, init=2)])
def test_client_with_extra_options(timeout: Union[Tuple[int, int], Timeout]) -> None:
    additional_config = AdditionalConfig(timeout=timeout, trust_env=True)
    client = weaviate.connect_to_embedded(
        port=8070,
        grpc_port=50040,
        additional_config=additional_config,
        environment_variables={"DISABLE_TELEMETRY": "true"},
    )
    try:
        assert client._connection.timeout_config == Timeout(query=1, insert=2, init=2)
    finally:
        client.close()


def test_connect_and_close_to_embedded() -> None:
    # Can't use the default port values as they are already in use by the local instances
    client = weaviate.connect_to_embedded(
        port=8078, grpc_port=50151, environment_variables={"DISABLE_TELEMETRY": "true"}
    )
    try:
        assert client.is_connected()
        metadata = client.get_meta()
        assert "1.23.7" == metadata["version"]
        assert client.is_ready()
        assert "8078" == metadata["hostname"].split(":")[2]
        assert client.is_live()

        client.close()
        assert not client.is_connected()
        with pytest.raises(WeaviateClosedClientError):
            client.get_meta()
    finally:
        client.close()


def test_embedded_as_context_manager() -> None:
    default_version = "1.23.7"
    with weaviate.connect_to_embedded(
        port=8077, grpc_port=50152, environment_variables={"DISABLE_TELEMETRY": "true"}
    ) as client:
        assert client.is_connected()
        metadata = client.get_meta()
        assert default_version == metadata["version"]
        assert client.is_ready()
        assert client.is_live()

    assert not client.is_connected()
    with pytest.raises(WeaviateClosedClientError):
        client.get_meta()


def test_embedded_with_wrong_version() -> None:
    with pytest.raises(weaviate.exceptions.WeaviateEmbeddedInvalidVersionError):
        weaviate.connect_to_embedded(
            version="this_version_does_not_exist",
            environment_variables={"DISABLE_TELEMETRY": "true"},
        )


def test_embedded_already_running() -> None:
    client = weaviate.connect_to_embedded(
        port=8096, grpc_port=50155, environment_variables={"DISABLE_TELEMETRY": "true"}
    )
    try:
        assert client._connection.embedded_db is not None
        assert client._connection.embedded_db.process is not None

        with pytest.raises(weaviate.exceptions.WeaviateStartUpError):
            weaviate.connect_to_embedded(
                port=8096, grpc_port=50155, environment_variables={"DISABLE_TELEMETRY": "true"}
            )
    finally:
        client.close()


def test_embedded_startup_with_blocked_http_port() -> None:
    client = weaviate.connect_to_embedded(
        port=8098, grpc_port=50096, environment_variables={"DISABLE_TELEMETRY": "true"}
    )
    try:
        with pytest.raises(weaviate.exceptions.WeaviateStartUpError):
            weaviate.connect_to_embedded(
                port=8098, grpc_port=50097, environment_variables={"DISABLE_TELEMETRY": "true"}
            )
    finally:
        client.close()


def test_embedded_startup_with_blocked_grpc_port() -> None:
    client = weaviate.connect_to_embedded(
        port=8099, grpc_port=50150, environment_variables={"DISABLE_TELEMETRY": "true"}
    )
    try:
        with pytest.raises(weaviate.exceptions.WeaviateStartUpError):
            weaviate.connect_to_embedded(
                port=8100, grpc_port=50150, environment_variables={"DISABLE_TELEMETRY": "true"}
            )
    finally:
        client.close()



# ORM API is not yet supported

# from typing import Type, Optional, Any, Dict, Generic, Tuple

# from pydantic import create_model

# from weaviate.collections.base import _CollectionBase, _CollectionsBase
# from weaviate.collections.classes.config import ConsistencyLevel
# from weaviate.collections.classes.orm import (
#     BaseProperty,
#     CollectionModelConfig,
#     Model,
#     UserModelType,
# )
# from weaviate.collections.config import _ConfigCollectionModel
# from weaviate.collections.data import _DataCollectionModel

# # from weaviate.collections.query import _GrpcCollectionModel
# from weaviate.collections.tenants import _Tenants
# from weaviate.connect import ConnectionV4
# from weaviate.exceptions import UnexpectedStatusCodeError
# from weaviate.util import _capitalize_first_letter
# from weaviate.types import PYTHON_TYPE_TO_DATATYPE


# class _CollectionObjectModel(_CollectionBase, Generic[Model]):
#     def __init__(
#         self,
#         connection: ConnectionV4,
#         name: str,
#         model: Type[Model],
#         config: _ConfigCollectionModel,
#         consistency_level: Optional[ConsistencyLevel] = None,
#         tenant: Optional[str] = None,
#     ) -> None:
#         super().__init__(connection, name)

#         self.config = config
#         self.data = _DataCollectionModel[Model](
#             connection, self.name, model, consistency_level, tenant
#         )
#         # self.query = _GrpcCollectionModel[Model](
#         #     connection, self.name, model, tenant, consistency_level
#         # )
#         self.tenants = _Tenants(connection, self.name)

#         self.__consistency_level = consistency_level
#         self.__model: Type[Model] = model
#         self.__tenant = tenant

#     @property
#     def model(self) -> Type[Model]:
#         return self.__model

#     def with_tenant(self, tenant: Optional[str] = None) -> "_CollectionObjectModel[Model]":
#         return _CollectionObjectModel[Model](
#             self._connection, self.name, self.__model, self.config, self.__consistency_level, tenant
#         )

#     def with_consistency_level(
#         self, consistency_level: Optional[ConsistencyLevel] = None
#     ) -> "_CollectionObjectModel[Model]":
#         return _CollectionObjectModel[Model](
#             self._connection, self.name, self.__model, self.config, consistency_level, self.__tenant
#         )


# class _CollectionModel(_CollectionsBase):
#     def create(self, config: CollectionModelConfig[Model]) -> _CollectionObjectModel[Model]:
#         name = super()._create(config._to_dict())
#         config_name = _capitalize_first_letter(config.model.__name__)
#         if config_name != name:
#             raise ValueError(
#                 f"Name of created collection ({name}) does not match given name ({config_name})"
#             )
#         return self.get(config.model)

#     def get(self, model: Type[Model]) -> _CollectionObjectModel[Model]:
#         name = _capitalize_first_letter(model.__name__)
#         config = _ConfigCollectionModel(self._connection, name, None)
#         if config.is_invalid(model):
#             raise TypeError(
#                 f"Model {model.__name__} definition does not match collection {name} config"
#             )
#         return _CollectionObjectModel[Model](self._connection, name, model, config)

#     def get_dynamic(self, name: str) -> Tuple[_CollectionObjectModel[BaseProperty], UserModelType]:
#         path = f"/schema/{_capitalize_first_letter(name)}"

#         response = self._connection.get(path=path, error_msg="Collection could not be retrieved.")
#         if response.status_code != 200:
#             raise UnexpectedStatusCodeError("Get schema", response)

#         response_json = response.json()
#         fields: Dict[str, Any] = {
#             prop["name"]: (PYTHON_TYPE_TO_DATATYPE[prop["dataType"][0]], None)
#             for prop in response_json["properties"]
#         }
#         model = create_model(response_json["class"], **fields, __base__=BaseProperty)
#         config = _ConfigCollectionModel(self._connection, name, None)
#         return _CollectionObjectModel[BaseProperty](self._connection, name, model, config), model

#     def delete(self, model: Type[Model]) -> None:
#         """Use this method to delete a collection from the Weaviate instance by its ORM model.

#         WARNING: If you have instances of client.orm.get() or client.orm.create()
#         for this collection within your code, they will cease to function correctly after this operation.

#         Parameters:
#             - model: The ORM model of the collection to be deleted.
#         """
#         name = _capitalize_first_letter(model.__name__)
#         return self._delete(name)

#     def exists(self, model: Type[Model]) -> bool:
#         name = _capitalize_first_letter(model.__name__)
#         return self._exists(name)

#     def update(self, model: Type[Model]) -> _CollectionObjectModel[Model]:
#         name = _capitalize_first_letter(model.__name__)
#         config = _ConfigCollectionModel(self._connection, name, None)
#         config.update_model(model)
#         return _CollectionObjectModel[Model](self._connection, name, model, config)

# class _DataCollectionModel(Generic[Model], _Data):
#     def __init__(
#         self,
#         connection: ConnectionV4,
#         name: str,
#         model: Type[Model],
#         consistency_level: Optional[ConsistencyLevel],
#         tenant: Optional[str],
#     ):
#         super().__init__(connection, name, consistency_level, tenant)
#         self.__model = model

#     def _json_to_object(self, obj: Dict[str, Any]) -> Object[Model, dict]:
#         for ref in self.__model.get_ref_fields(self.__model):
#             if ref not in obj["properties"]:
#                 continue

#             beacons = obj["properties"][ref]
#             uuids = []
#             for beacon in beacons:
#                 uri = beacon["beacon"]
#                 assert isinstance(uri, str)
#                 uuids.append(uri.split("/")[-1])

#             obj["properties"][ref] = uuids

#         # weaviate does not save none values, so we need to add them to pass model validation
#         for prop in self.__model.get_non_default_fields(self.__model):
#             if prop not in obj["properties"]:
#                 obj["properties"][prop] = None

#         # uuid, vector, metadata = _metadata_from_dict(obj)
#         uuid = uuid_package.uuid4()
#         metadata = MetadataReturn()
#         model_object = Object[Model, dict](
#             collection=self.name,
#             properties=self.__model.model_validate(
#                 {
#                     **obj["properties"],
#                     # "uuid": uuid,
#                     # "vector": vector,
#                 }
#             ),
#             references={},
#             metadata=metadata,
#             uuid=uuid,
#             vector=None,
#         )
#         return model_object

#     def insert(self, obj: Model) -> uuid_package.UUID:
#         self.__model.model_validate(obj)
#         weaviate_obj: Dict[str, Any] = {
#             "class": self.name,
#             "properties": self._serialize_props(obj.props_to_dict()),
#             "id": str(obj.uuid),
#         }
#         if obj.vector is not None:
#             weaviate_obj["vector"] = obj.vector

#         self._insert(weaviate_obj)
#         return uuid_package.UUID(str(obj.uuid))

#     def insert_many(self, objects: List[Model]) -> BatchObjectReturn:
#         for obj in objects:
#             self.__model.model_validate(obj)

#         data_objects = [
#             _BatchObject(
#                 collection=self.name,
#                 properties=obj.props_to_dict(),
#                 tenant=self._tenant,
#                 uuid=obj.uuid,
#                 vector=obj.vector,
#                 references=None,
#             )
#             for obj in objects
#         ]

#         return self._batch_grpc.objects(data_objects, self._connection.timeout_config.connect)

#     def replace(self, obj: Model, uuid: UUID) -> None:
#         self.__model.model_validate(obj)

#         weaviate_obj: Dict[str, Any] = {
#             "class": self.name,
#             "properties": self._serialize_props(obj.props_to_dict()),
#         }
#         if obj.vector is not None:
#             weaviate_obj["vector"] = obj.vector

#         self._replace(weaviate_obj, uuid)

#     def update(self, obj: Model, uuid: UUID) -> None:
#         self.__model.model_validate(obj)

#         weaviate_obj: Dict[str, Any] = {
#             "class": self.name,
#             "properties": self._serialize_props(obj.props_to_dict()),
#         }
#         if obj.vector is not None:
#             weaviate_obj["vector"] = obj.vector

#         self._update(weaviate_obj, uuid)

#     def reference_add(self, from_uuid: UUID, from_property: str, ref: _Reference) -> None:
#         self._reference_add(from_uuid=from_uuid, from_property=from_property, ref=ref)

#     def reference_delete(self, from_uuid: UUID, from_property: str, ref: _Reference) -> None:
#         self._reference_delete(from_uuid=from_uuid, from_property=from_property, ref=ref)

#     def reference_replace(self, from_uuid: UUID, from_property: str, ref: _Reference) -> None:
#         self._reference_replace(from_uuid=from_uuid, from_property=from_property, ref=ref)

#     def reference_add_many(self, refs: List[DataReferences]) -> BatchReferenceReturn:
#         return self._reference_add_many(refs)

# class _ConfigCollectionModel(_ConfigBase):
#     def __compare_properties_with_model(
#         self, schema_props: List[_Property], model_props: List[PropertyType]
#     ) -> Tuple[List[_Property], List[PropertyType]]:
#         only_in_model: List[PropertyType] = []
#         only_in_schema: List[_Property] = list(schema_props)

#         schema_props_simple = [
#             {
#                 "name": prop.name,
#                 "dataType": prop.to_dict().get("dataType"),
#             }
#             for prop in schema_props
#         ]

#         for prop in model_props:
#             try:
#                 idx = schema_props_simple.index(
#                     {"name": prop.name, "dataType": prop._to_dict().get("dataType")}
#                 )
#                 schema_props_simple.pop(idx)
#                 only_in_schema.pop(idx)
#             except ValueError:
#                 only_in_model.append(prop)
#         return only_in_schema, only_in_model

#     def update_model(self, model: Type[Model]) -> None:
#         only_in_schema, only_in_model = self.__compare_properties_with_model(
#             self.get().properties, model.type_to_properties(model)
#         )
#         if len(only_in_schema) > 0:
#             raise TypeError("Schema has extra properties")

#         # we can only allow new optional types unless the default is None
#         for prop in only_in_model:
#             new_field = model.model_fields[prop.name]
#             if new_field.annotation is None:
#                 continue  # if user did not annotate with type then ignore field
#             non_optional_type = model.remove_optional_type(new_field.annotation)
#             if new_field.default is not None and non_optional_type == new_field.annotation:
#                 raise WeaviateAddInvalidPropertyError(prop.name)

#         for prop in only_in_model:
#             self._add_property(prop)

#     def is_invalid(self, model: Type[Model]) -> bool:
#         only_in_schema, only_in_model = self.__compare_properties_with_model(
#             self.get().properties, model.type_to_properties(model)
#         )
#         return len(only_in_schema) > 0 or len(only_in_model) > 0



from typing import Generator, Tuple, Union

import pytest
from _pytest.fixtures import SubRequest

import weaviate
from weaviate.collections import Collection
from weaviate.collections.classes.config import (
    Configure,
    _CollectionConfig,
    DataType,
    GenerativeSearches,
    Property,
    ReferenceProperty,
    Vectorizers,
)
from weaviate.exceptions import WeaviateClosedClientError, WeaviateStartUpError
import weaviate.classes as wvc

from weaviate.config import Timeout

WCS_HOST = "piblpmmdsiknacjnm1ltla.c1.europe-west3.gcp.weaviate.cloud"
WCS_URL = f"https://{WCS_HOST}"
WCS_GRPC_HOST = f"grpc-{WCS_HOST}"
WCS_CREDS = wvc.init.Auth.api_key("cy4ua772mBlMdfw3YnclqAWzFhQt0RLIN0sl")


@pytest.fixture(scope="module")
def client() -> Generator[weaviate.WeaviateClient, None, None]:
    client = weaviate.WeaviateClient(
        connection_params=weaviate.connect.ConnectionParams.from_url(
            "http://localhost:8080", 50051
        ),
        skip_init_checks=False,
    )
    client.connect()
    try:
        yield client
    finally:
        client.close()


def test_fail_to_connect_to_inactive_grpc_port() -> None:
    with pytest.raises(weaviate.exceptions.WeaviateGRPCUnavailableError):
        weaviate.WeaviateClient(
            connection_params=weaviate.connect.ConnectionParams.from_url(
                "http://localhost:8080", 12345
            ),
            skip_init_checks=False,
        ).connect()


def test_fail_to_connect_with_bad_wcs_url() -> None:
    with pytest.raises(WeaviateStartUpError):
        weaviate.connect_to_wcs(
            WCS_URL + "bad",
            auth_credentials=WCS_CREDS,
        ).connect()


@pytest.mark.parametrize(
    "bad_config",
    [
        {
            "http_secure": False,
            "http_host": WCS_HOST,
            "http_port": 443,
            "grpc_secure": True,
            "grpc_host": WCS_GRPC_HOST,
            "grpc_port": 443,
        },
        {
            "http_secure": True,
            "http_host": WCS_HOST,
            "http_port": 80,
            "grpc_secure": True,
            "grpc_host": WCS_GRPC_HOST,
            "grpc_port": 443,
        },
        {
            "http_secure": True,
            "http_host": WCS_HOST + "bad",
            "http_port": 443,
            "grpc_secure": True,
            "grpc_host": WCS_GRPC_HOST,
            "grpc_port": 443,
        },
    ],
)
def test_fail_to_connect_with_bad_custom_wcs_setup_rest(bad_config: dict) -> None:
    with pytest.raises(WeaviateStartUpError):
        weaviate.connect_to_custom(**bad_config, auth_credentials=WCS_CREDS)


@pytest.mark.parametrize(
    "bad_config",
    [
        {
            "http_secure": True,
            "http_host": WCS_HOST,
            "http_port": 443,
            "grpc_secure": False,
            "grpc_host": WCS_GRPC_HOST,
            "grpc_port": 443,
        },
        {
            "http_secure": True,
            "http_host": WCS_HOST,
            "http_port": 443,
            "grpc_secure": True,
            "grpc_host": WCS_GRPC_HOST + "bad",
            "grpc_port": 443,
        },
        {
            "http_secure": True,
            "http_host": WCS_HOST,
            "http_port": 443,
            "grpc_secure": True,
            "grpc_host": WCS_GRPC_HOST,
            "grpc_port": 80,
        },
    ],
)
def test_fail_to_connect_with_bad_custom_wcs_setup_grpc(bad_config: dict) -> None:
    with pytest.raises(weaviate.exceptions.WeaviateGRPCUnavailableError):
        weaviate.connect_to_custom(**bad_config, auth_credentials=WCS_CREDS)


def test_fail_to_connect_with_bad_custom_wcs_setup_rest_and_grpc() -> None:
    """Test that REST checks take precendence to gRPC checks by throwing REST error rather than gRPC error."""
    with pytest.raises(WeaviateStartUpError):
        weaviate.connect_to_custom(
            http_secure=True,
            http_host=WCS_HOST + "bad",
            http_port=443,
            grpc_secure=True,
            grpc_host=WCS_GRPC_HOST + "bad",
            grpc_port=443,
            auth_credentials=WCS_CREDS,
        )


def test_connect_to_wcs() -> None:
    with weaviate.connect_to_wcs(
        "https://piblpmmdsiknacjnm1ltla.c1.europe-west3.gcp.weaviate.cloud",
        auth_credentials=WCS_CREDS,
    ) as client:
        client.get_meta()


def test_create_get_and_delete(client: weaviate.WeaviateClient, request: SubRequest) -> None:
    name = request.node.name
    client.collections.delete(name)

    col = client.collections.create(name=name, vectorizer_config=Configure.Vectorizer.none())
    assert client.collections.exists(name)
    assert isinstance(col, Collection)

    col = client.collections.get(name)
    assert isinstance(col, Collection)

    client.collections.delete(name)
    assert not client.collections.exists(name)


def test_delete_multiple(client: weaviate.WeaviateClient, request: SubRequest) -> None:
    name1 = request.node.name
    name2 = request.node.name + "2"
    client.collections.delete([name1, name2])

    client.collections.create(
        name=name1,
        vectorizer_config=Configure.Vectorizer.none(),
    )
    client.collections.create(
        name=name2,
        vectorizer_config=Configure.Vectorizer.none(),
    )
    assert client.collections.exists(name1)
    assert client.collections.exists(name2)

    client.collections.delete([name1, name2])
    assert not client.collections.exists(name1)
    assert not client.collections.exists(name2)


def test_create_raw_get_and_delete(client: weaviate.WeaviateClient, request: SubRequest) -> None:
    name = request.node.name
    col = client.collections.create_from_dict(
        {
            "class": name,
            "properties": [{"name": "Name", "dataType": ["text"]}],
            "vectorizer": "none",
        }
    )
    assert client.collections.exists(name)
    assert isinstance(col, Collection)

    col = client.collections.get(name)
    assert isinstance(col, Collection)

    client.collections.delete(name)
    assert not client.collections.exists(name)


def test_create_export_and_recreate(client: weaviate.WeaviateClient, request: SubRequest) -> None:
    name1 = request.node.name
    name2 = request.node.name + "2"
    client.collections.delete([name1, name2])

    col = client.collections.create(
        name=name1,
        vectorizer_config=Configure.Vectorizer.text2vec_contextionary(
            vectorize_collection_name=False
        ),
        generative_config=Configure.Generative.cohere(model="something", k=10),
        properties=[
            Property(
                name="name",
                data_type=DataType.TEXT,
                description="desc",
                index_filterable=True,
                index_searchable=True,
                skip_vectorization=True,
                vectorize_property_name=True,
            ),
            Property(
                name="name_vectorized",
                data_type=DataType.TEXT,
                description="desc",
                index_filterable=True,
                index_searchable=True,
                skip_vectorization=False,
                vectorize_property_name=False,
            ),
        ],
        references=[ReferenceProperty(name="ref", target_collection=name1)],
    )
    assert client.collections.exists(name1)
    assert isinstance(col, Collection)

    export = client.collections.export_config(name1)
    assert isinstance(export, _CollectionConfig)
    export.name = name2

    col = client.collections.create_from_config(export)
    assert client.collections.exists(name2)
    assert isinstance(col, Collection)
    export = client.collections.export_config(name2)
    assert len(export.properties) == 2
    assert export.properties[0].description == "desc"
    assert len(export.references) == 1
    assert export.properties[0].index_searchable
    assert export.vectorizer_config is not None
    assert export.vectorizer_config.vectorizer == Vectorizers.TEXT2VEC_CONTEXTIONARY
    assert not export.vectorizer_config.vectorize_collection_name

    assert export.generative_config is not None
    assert export.generative_config.generative == GenerativeSearches.COHERE
    assert export.generative_config.model["model"] == "something"
    assert export.generative_config.model["kProperty"] == 10

    client.collections.delete([name1, name2])
    assert not client.collections.exists(name1)
    assert not client.collections.exists(name2)


def test_create_export_and_recreate_named_vectors(
    client: weaviate.WeaviateClient, request: SubRequest
) -> None:
    if client._connection._weaviate_version.is_lower_than(1, 24, 0):
        pytest.skip("Named vectors are not supported in versions lower than 1.24.0")

    name1 = request.node.name
    name2 = request.node.name + "2"
    client.collections.delete([name1, name2])

    col = client.collections.create(
        name=name1,
        properties=[
            Property(
                name="name",
                data_type=DataType.TEXT,
                vectorize_property_name=True,
            ),
        ],
        vectorizer_config=[
            Configure.NamedVectors.text2vec_contextionary(
                "name",
                source_properties=["name"],
                vectorize_collection_name=False,
            ),
            Configure.NamedVectors.none("custom", vector_index_config=Configure.VectorIndex.flat()),
        ],
    )
    conf = col.config.get()
    conf.name = name2

    col2 = client.collections.create_from_config(conf)

    conf2 = col2.config.get()
    assert conf2.vector_config == conf.vector_config

    client.collections.delete([name1, name2])


def test_collection_name_capitalization(
    client: weaviate.WeaviateClient, request: SubRequest
) -> None:
    name_small = "collectionCapitalizationTest"
    name_big = "CollectionCapitalizationTest"
    try:
        collection = client.collections.create(
            name=name_small,
            vectorizer_config=Configure.Vectorizer.none(),
        )
        assert collection.name == name_big
        client.collections.delete(name_small)
        assert not client.collections.exists(name_small)
        assert not client.collections.exists(name_big)
    finally:
        client.collections.delete(name_big)


def test_client_cluster(client: weaviate.WeaviateClient, request: SubRequest) -> None:
    try:
        collection = client.collections.create(
            name=request.node.name, vectorizer_config=Configure.Vectorizer.none()
        )

        nodes = client.cluster.nodes(collection.name, output="verbose")
        assert len(nodes) == 1
        assert len(nodes[0].shards) == 1
        assert nodes[0].shards[0].collection == collection.name
        assert nodes[0].shards[0].object_count == 0
        assert nodes[0].shards[0].vector_indexing_status == "READY"
        assert nodes[0].shards[0].vector_queue_length == 0
        assert nodes[0].shards[0].compressed is False
        if collection._connection._weaviate_version.is_lower_than(1, 24, 0):
            assert nodes[0].shards[0].loaded is None
        else:
            assert nodes[0].shards[0].loaded is True
    finally:
        client.collections.delete(request.node.name)


def test_client_cluster_multitenant(client: weaviate.WeaviateClient, request: SubRequest) -> None:
    try:
        collection = client.collections.create(
            name=request.node.name,
            multi_tenancy_config=Configure.multi_tenancy(enabled=True),
            vectorizer_config=Configure.Vectorizer.none(),
        )

        nodes = client.cluster.nodes(collection.name, output="verbose")
        assert len(nodes) == 1
        assert len(nodes[0].shards) == 0
    finally:
        client.collections.delete(request.node.name)


def test_client_cluster_minimal(client: weaviate.WeaviateClient, request: SubRequest) -> None:
    try:
        collection = client.collections.create(
            name=request.node.name, vectorizer_config=Configure.Vectorizer.none()
        )

        nodes = client.cluster.nodes(collection.name, output="minimal")
        assert len(nodes) == 1
        assert nodes[0].shards is None
    finally:
        client.collections.delete(request.node.name)


def test_client_connect_and_close() -> None:
    client = weaviate.WeaviateClient(
        connection_params=weaviate.connect.ConnectionParams.from_url(
            "http://localhost:8080", 50051
        ),
        skip_init_checks=False,
    )
    client.connect()
    assert client.is_connected()
    client.get_meta()
    assert client.is_live()

    client.close()
    assert not client.is_connected()
    with pytest.raises(WeaviateClosedClientError):
        client.get_meta()


@pytest.mark.asyncio
async def test_async_client_connect_and_close() -> None:
    client = weaviate.WeaviateAsyncClient(
        connection_params=weaviate.connect.ConnectionParams.from_url(
            "http://localhost:8080", 50051
        ),
        skip_init_checks=False,
    )
    await client.connect()
    assert client.is_connected()
    await client.get_meta()
    assert await client.is_live()

    await client.close()
    assert not client.is_connected()
    with pytest.raises(WeaviateClosedClientError):
        await client.get_meta()


def test_client_as_context_manager() -> None:
    with weaviate.WeaviateClient(
        connection_params=weaviate.connect.ConnectionParams.from_url(
            "http://localhost:8080", 50051
        ),
        skip_init_checks=False,
    ) as client:
        assert client.is_connected()
        client.get_meta()
        assert client.is_live()

    assert not client.is_connected()
    with pytest.raises(WeaviateClosedClientError):
        client.get_meta()


@pytest.mark.asyncio
async def test_async_client_as_context_manager() -> None:
    async with weaviate.WeaviateAsyncClient(
        connection_params=weaviate.connect.ConnectionParams.from_url(
            "http://localhost:8080", 50051
        ),
        skip_init_checks=False,
    ) as client:
        assert client.is_connected()
        await client.get_meta()
        assert await client.is_live()

    assert not client.is_connected()
    with pytest.raises(WeaviateClosedClientError):
        await client.get_meta()


def test_connect_to_wrong_wcs() -> None:
    with pytest.raises(WeaviateStartUpError):
        weaviate.connect_to_wcs("does-not-exist", auth_credentials=WCS_CREDS, skip_init_checks=True)


def test_connect_to_wrong_local() -> None:
    with pytest.raises(expected_exception=WeaviateStartUpError):
        weaviate.connect_to_local("does-not-exist", skip_init_checks=True)


def test_connect_to_wrong_custom() -> None:
    with pytest.raises(expected_exception=WeaviateStartUpError):
        weaviate.connect_to_custom(
            "does-not-exist",
            http_port=1234,
            http_secure=False,
            grpc_host="does-not-exist",
            grpc_port=2345,
            grpc_secure=False,
            skip_init_checks=True,
        )


def test_rest_call_without_connect() -> None:
    client = weaviate.WeaviateClient(
        weaviate.connect.ConnectionParams.from_url("http://localhost:8080", 50051)
    )
    with pytest.raises(weaviate.exceptions.WeaviateClosedClientError):
        client.get_meta()


def test_grpc_call_without_connect() -> None:
    client = weaviate.WeaviateClient(
        weaviate.connect.ConnectionParams.from_url("http://localhost:8080", 50051)
    )
    with pytest.raises(weaviate.exceptions.WeaviateClosedClientError):
        client.collections.get("does-not-exist").query.fetch_objects()


def test_client_with_skip_init_check(request: SubRequest) -> None:
    with weaviate.connect_to_local(skip_init_checks=True) as client:
        client.collections.delete(request.node.name)
        col = client.collections.create(
            name=request.node.name,
            vectorizer_config=Configure.Vectorizer.none(),
            properties=[
                Property(name="name", data_type=DataType.TEXT),
                Property(name="age", data_type=DataType.INT),
            ],
        )

        col.data.insert(properties={"name": "Name"})

        obj = col.query.fetch_objects().objects[0]
        assert obj.properties["name"] == "Name"
        client.collections.delete(request.node.name)


@pytest.mark.parametrize("timeout", [(1, 2), Timeout(query=1, insert=2, init=2)])
def test_client_with_extra_options(timeout: Union[Tuple[int, int], Timeout]) -> None:
    additional_config = wvc.init.AdditionalConfig(timeout=timeout, trust_env=True)

    for client in [
        weaviate.connect_to_wcs(
            cluster_url=WCS_URL, auth_credentials=WCS_CREDS, additional_config=additional_config
        ),
        weaviate.connect_to_weaviate_cloud(
            cluster_url=WCS_URL, auth_credentials=WCS_CREDS, additional_config=additional_config
        ),
        weaviate.connect_to_local(additional_config=additional_config),
        weaviate.connect_to_custom(
            http_secure=True,
            http_host=WCS_HOST,
            http_port=443,
            grpc_secure=True,
            grpc_host=WCS_GRPC_HOST,
            grpc_port=443,
            auth_credentials=WCS_CREDS,
            additional_config=additional_config,
        ),
    ]:
        assert client._connection.timeout_config == Timeout(query=1, insert=2, init=2)
        client.close()


@pytest.mark.parametrize("timeout", [(1, 2), Timeout(query=1, insert=2, init=2)])
@pytest.mark.asyncio
async def test_async_client_with_extra_options(timeout: Union[Tuple[int, int], Timeout]) -> None:
    additional_config = wvc.init.AdditionalConfig(timeout=timeout, trust_env=True)

    for client in [
        weaviate.use_async_with_weaviate_cloud(
            cluster_url=WCS_URL,
            auth_credentials=WCS_CREDS,
            additional_config=additional_config,
        ),
        weaviate.use_async_with_local(additional_config=additional_config),
        weaviate.use_async_with_custom(
            http_secure=True,
            http_host=WCS_HOST,
            http_port=443,
            grpc_secure=True,
            grpc_host=WCS_GRPC_HOST,
            grpc_port=443,
            auth_credentials=WCS_CREDS,
            additional_config=additional_config,
        ),
    ]:
        await client.connect()
        await client.get_meta()
        assert client._connection.timeout_config == Timeout(query=1, insert=2, init=2)
        await client.close()


def test_client_error_for_wcs_without_auth() -> None:
    with pytest.raises(weaviate.exceptions.AuthenticationFailedError) as e:
        weaviate.connect_to_wcs(cluster_url=WCS_URL, auth_credentials=None)
        assert "wvc.init.Auth.api_key" in e.value.message


def test_client_is_not_ready() -> None:
    assert not weaviate.WeaviateClient(
        connection_params=weaviate.connect.ConnectionParams.from_url(
            "http://localhost:8080", 50051
        ),
        skip_init_checks=True,
    ).is_ready()


def test_client_is_ready() -> None:
    with weaviate.connect_to_wcs(
        cluster_url=WCS_URL, auth_credentials=WCS_CREDS, skip_init_checks=True
    ) as client:
        assert client.is_ready()


@pytest.mark.asyncio
async def test_async_client_is_ready() -> None:
    async with weaviate.use_async_with_weaviate_cloud(
        cluster_url=WCS_URL, auth_credentials=WCS_CREDS, skip_init_checks=True
    ) as client:
        assert await client.is_ready()


@pytest.mark.skip("gRPC proxying is not supported by grpclib at this time")
def test_local_proxies() -> None:
    with weaviate.connect_to_local(
        additional_config=wvc.init.AdditionalConfig(
            proxies=wvc.init.Proxies(
                http="http://localhost:8075",
                grpc="http://localhost:10000",
            )
        )
    ) as client:
        client.collections.delete("TestLocalProxies")
        collection = client.collections.create(
            "TestLocalProxies",
            properties=[wvc.config.Property(name="name", data_type=wvc.config.DataType.TEXT)],
            vectorizer_config=Configure.Vectorizer.none(),
        )
        collection.data.insert({"name": "Test"})
        assert collection.query.fetch_objects().objects[0].properties["name"] == "Test"



from typing import Any, List

import numpy as np
import pandas as pd
import polars as pl
import pytest

from weaviate.exceptions import WeaviateInvalidInputError
from weaviate.validator import _validate_input, _ValidateArgument, _ExtraTypes


@pytest.mark.parametrize(
    "inputs,expected,error",
    [
        (1, [int], False),
        (1.0, [int], True),
        ([1, 1], [List], False),
        (np.array([1, 2, 3]), [_ExtraTypes.NUMPY], False),
        (np.array([1, 2, 3]), [_ExtraTypes.NUMPY, List], False),
        (np.array([1, 2, 3]), [List], True),
        ([1, 1], [List, _ExtraTypes.NUMPY], False),
        (pd.array([1, 1]), [_ExtraTypes.PANDAS, List], False),
        (pd.Series([1, 1]), [_ExtraTypes.PANDAS, List], False),
        (pl.Series([1, 1]), [_ExtraTypes.POLARS, List], False),
        (
            pl.Series([1, 1]),
            [_ExtraTypes.POLARS, _ExtraTypes.PANDAS, _ExtraTypes.NUMPY, List],
            False,
        ),
        (pl.Series([1, 1]), [_ExtraTypes.PANDAS, _ExtraTypes.NUMPY, List], True),
    ],
)
def test_validator(inputs: Any, expected: List[Any], error: bool) -> None:
    if error:
        with pytest.raises(WeaviateInvalidInputError):
            _validate_input(_ValidateArgument(expected=expected, name="test", value=inputs))
    else:
        _validate_input(_ValidateArgument(expected=expected, name="test", value=inputs))



from .generate import _NearTextGenerateAsync, _NearTextGenerate
from .query import _NearTextQueryAsync, _NearTextQuery

__all__ = [
    "_NearTextGenerate",
    "_NearTextQuery",
    "_NearTextGenerateAsync",
    "_NearTextQueryAsync",
]



import numpy as np
import pandas as pd
import polars as pl
from weaviate.util import _get_vector_v4

# Don't need to test `tf` or `torch` tensors since `_get_vector_v4`` converts them to `np` first


def test_get_vector_v4_with_np_array_and_np_float() -> None:
    # Assert that np.array[np.float32] -> list[float]
    vec = _get_vector_v4(np.array([1.0, 2.0, 3.0, 4.0, 5.0], dtype=np.float32))
    assert isinstance(vec, list)
    assert all(isinstance(e, float) for e in vec)


def test_get_vector_v4_with_list_and_np_float() -> None:
    # Assert that list[np.float32] -> list[np.float32]
    vec = _get_vector_v4(list(np.array([0.12345] * 1536, dtype=np.float32)))
    assert isinstance(vec, list)
    assert all(isinstance(e, np.float32) for e in vec)


def test_get_vector_v4_with_list() -> None:
    # Assert that list[float] -> list[float]
    vec = _get_vector_v4([1.0, 2.0, 3.0, 4.0, 5.0])
    assert isinstance(vec, list)
    assert all(isinstance(e, float) for e in vec)


def test_get_vector_v4_with_pd() -> None:
    # Assert that pd.Series[pd.Float32Dtype] -> list[float]
    vec = _get_vector_v4(pd.Series([1.0, 2.0, 3.0, 4.0, 5.0], dtype="Float32"))
    assert isinstance(vec, list)
    assert all(isinstance(e, float) for e in vec)


def test_get_vector_v4_with_pl() -> None:
    # Assert that pl.Series[pl.Float32] -> list[float]
    vec = _get_vector_v4(pl.Series("a", [1.0, 2.0, 3.0, 4.0, 5.0], pl.Float32))
    assert isinstance(vec, list)
    assert all(isinstance(e, float) for e in vec)



from dataclasses import dataclass
from enum import Enum, auto
from typing import ClassVar, List, Literal, Optional, Sequence, Type, Union, Dict

from pydantic import ConfigDict, Field

from weaviate.collections.classes.types import _WeaviateInput
from weaviate.proto.v1 import search_get_pb2
from weaviate.str_enum import BaseEnum
from weaviate.types import INCLUDE_VECTOR, UUID


class HybridFusion(str, BaseEnum):
    """Define how the query's hybrid fusion operation should be performed."""

    RANKED = "FUSION_TYPE_RANKED"
    RELATIVE_SCORE = "FUSION_TYPE_RELATIVE_SCORE"


class Move:
    """Define how the query's move operation should be performed."""

    def __init__(
        self,
        force: float,
        objects: Optional[Union[List[UUID], UUID]] = None,
        concepts: Optional[Union[List[str], str]] = None,
    ):
        if (objects is None or (isinstance(objects, list) and len(objects) == 0)) and (
            concepts is None or (isinstance(concepts, list) and len(concepts) == 0)
        ):
            raise ValueError("Either objects or concepts need to be given")

        self.force = force

        # accept single values, but make them a list
        if objects is None:
            self.__objects = None
        elif not isinstance(objects, list):
            self.__objects = [str(objects)]
        else:
            self.__objects = [str(obj_uuid) for obj_uuid in objects]

        if concepts is None:
            self.__concepts = None
        elif not isinstance(concepts, list):
            self.__concepts = [concepts]
        else:
            self.__concepts = concepts

    @property
    def _objects_list(self) -> Optional[List[str]]:
        return self.__objects

    @property
    def _concepts_list(self) -> Optional[List[str]]:
        return self.__concepts

    def _to_gql_payload(self) -> dict:
        payload: dict = {"force": self.force}
        if self.__objects is not None:
            payload["objects"] = [{"id": obj} for obj in self.__objects]
        if self.__concepts is not None:
            payload["concepts"] = self.__concepts
        return payload


class MetadataQuery(_WeaviateInput):
    """Define which metadata should be returned in the query's results."""

    creation_time: bool = Field(default=False)
    last_update_time: bool = Field(default=False)
    distance: bool = Field(default=False)
    certainty: bool = Field(default=False)
    score: bool = Field(default=False)
    explain_score: bool = Field(default=False)
    is_consistent: bool = Field(default=False)

    @classmethod
    def full(cls) -> "MetadataQuery":
        """Return a MetadataQuery with all fields set to True."""
        return cls(
            creation_time=True,
            last_update_time=True,
            distance=True,
            certainty=True,
            score=True,
            explain_score=True,
            is_consistent=True,
        )


@dataclass
class _MetadataQuery:
    vector: bool
    uuid: bool = True
    creation_time_unix: bool = False
    last_update_time_unix: bool = False
    distance: bool = False
    certainty: bool = False
    score: bool = False
    explain_score: bool = False
    is_consistent: bool = False
    vectors: Optional[List[str]] = None

    @classmethod
    def from_public(
        cls, public: Optional[MetadataQuery], include_vector: INCLUDE_VECTOR
    ) -> "_MetadataQuery":
        return (
            cls(
                vector=include_vector if isinstance(include_vector, bool) else False,
                vectors=include_vector if isinstance(include_vector, list) else None,
            )
            if public is None
            else cls(
                vector=include_vector if isinstance(include_vector, bool) else False,
                vectors=include_vector if isinstance(include_vector, list) else None,
                creation_time_unix=public.creation_time,
                last_update_time_unix=public.last_update_time,
                distance=public.distance,
                certainty=public.certainty,
                score=public.score,
                explain_score=public.explain_score,
                is_consistent=public.is_consistent,
            )
        )


METADATA = Union[
    List[
        Literal[
            "creation_time",
            "last_update_time",
            "distance",
            "certainty",
            "score",
            "explain_score",
            "is_consistent",
        ]
    ],
    MetadataQuery,
]


class Generate(_WeaviateInput):
    """Define how the query's RAG capabilities should be performed."""

    single_prompt: Optional[str] = Field(default=None)
    grouped_task: Optional[str] = Field(default=None)
    grouped_properties: Optional[List[str]] = Field(default=None)


class GroupBy(_WeaviateInput):
    """Define how the query's group-by operation should be performed."""

    prop: str
    objects_per_group: int
    number_of_groups: int


class _Sort(_WeaviateInput):
    prop: str
    ascending: bool = Field(default=True)


class _Sorting:
    def __init__(self) -> None:
        self.sorts: List[_Sort] = []

    def by_property(self, name: str, ascending: bool = True) -> "_Sorting":
        """Sort by an object property in the collection."""
        self.sorts.append(_Sort(prop=name, ascending=ascending))
        return self

    def by_id(self, ascending: bool = True) -> "_Sorting":
        """Sort by an object's ID in the collection."""
        self.sorts.append(_Sort(prop="_id", ascending=ascending))
        return self

    def by_creation_time(self, ascending: bool = True) -> "_Sorting":
        """Sort by an object's creation time."""
        self.sorts.append(_Sort(prop="_creationTimeUnix", ascending=ascending))
        return self

    def by_update_time(self, ascending: bool = True) -> "_Sorting":
        """Sort by an object's last update time."""
        self.sorts.append(_Sort(prop="_lastUpdateTimeUnix", ascending=ascending))
        return self


Sorting = _Sorting
"""The type returned by the `Sort` class to be used when defining programmatic sort chains."""


class Sort:
    """Define how the query's sort operation should be performed using the available static methods."""

    def __init__(self) -> None:
        raise TypeError("Sort cannot be instantiated. Use the static methods to create a sorter.")

    @staticmethod
    def by_property(name: str, ascending: bool = True) -> Sorting:
        """Sort by an object property in the collection."""
        return _Sorting().by_property(name=name, ascending=ascending)

    @staticmethod
    def by_id(ascending: bool = True) -> Sorting:
        """Sort by an object's ID in the collection."""
        return _Sorting().by_id(ascending=ascending)

    @staticmethod
    def by_creation_time(ascending: bool = True) -> Sorting:
        """Sort by an object's creation time."""
        return _Sorting().by_creation_time(ascending=ascending)

    @staticmethod
    def by_update_time(ascending: bool = True) -> Sorting:
        """Sort by an object's last update time."""
        return _Sorting().by_update_time(ascending=ascending)


class Rerank(_WeaviateInput):
    """Define how the query's rerank operation should be performed."""

    prop: str
    query: Optional[str] = Field(default=None)


class _HybridNearBase(_WeaviateInput):
    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid")

    distance: Optional[float] = None
    certainty: Optional[float] = None


class _HybridNearText(_HybridNearBase):
    text: Union[str, List[str]]
    move_to: Optional[Move] = None
    move_away: Optional[Move] = None


class _HybridNearVector(_HybridNearBase):
    vector: List[float]


HybridVectorType = Union[List[float], _HybridNearText, _HybridNearVector]


class _MultiTargetVectorJoinEnum(BaseEnum):
    """Define how multi target vector searches should be combined."""

    SUM = auto()
    AVERAGE = auto()
    MINIMUM = auto()
    RELATIVE_SCORE = auto()
    MANUAL_WEIGHTS = auto()


@dataclass
class _MultiTargetVectorJoin:
    combination: _MultiTargetVectorJoinEnum
    target_vectors: List[str]
    weights: Optional[Dict[str, float]] = None

    def to_grpc_target_vector(self) -> search_get_pb2.Targets:
        combination = self.combination
        if combination == _MultiTargetVectorJoinEnum.AVERAGE:
            combination_grpc = search_get_pb2.COMBINATION_METHOD_TYPE_AVERAGE
        elif combination == _MultiTargetVectorJoinEnum.SUM:
            combination_grpc = search_get_pb2.COMBINATION_METHOD_TYPE_SUM
        elif combination == _MultiTargetVectorJoinEnum.RELATIVE_SCORE:
            combination_grpc = search_get_pb2.COMBINATION_METHOD_TYPE_RELATIVE_SCORE
        elif combination == _MultiTargetVectorJoinEnum.MANUAL_WEIGHTS:
            combination_grpc = search_get_pb2.COMBINATION_METHOD_TYPE_MANUAL
        else:
            assert combination == _MultiTargetVectorJoinEnum.MINIMUM
            combination_grpc = search_get_pb2.COMBINATION_METHOD_TYPE_MIN

        return search_get_pb2.Targets(
            target_vectors=self.target_vectors, weights=self.weights, combination=combination_grpc
        )


TargetVectorJoinType = Union[str, List[str], _MultiTargetVectorJoin]


class TargetVectors:
    """Define how the distances from different target vectors should be combined using the available methods."""

    @staticmethod
    def sum(target_vectors: List[str]) -> _MultiTargetVectorJoin:  # noqa: A003
        """Combine the distance from different target vectors by summing them."""
        return _MultiTargetVectorJoin(
            combination=_MultiTargetVectorJoinEnum.SUM, target_vectors=target_vectors
        )

    @staticmethod
    def average(target_vectors: List[str]) -> _MultiTargetVectorJoin:
        """Combine the distance from different target vectors by averaging them."""
        return _MultiTargetVectorJoin(
            combination=_MultiTargetVectorJoinEnum.AVERAGE, target_vectors=target_vectors
        )

    @staticmethod
    def minimum(target_vectors: List[str]) -> _MultiTargetVectorJoin:
        """Combine the distance from different target vectors by using the minimum distance."""
        return _MultiTargetVectorJoin(
            combination=_MultiTargetVectorJoinEnum.MINIMUM, target_vectors=target_vectors
        )

    @staticmethod
    def manual_weights(weights: Dict[str, float]) -> _MultiTargetVectorJoin:
        """Combine the distance from different target vectors by summing them using manual weights."""
        return _MultiTargetVectorJoin(
            combination=_MultiTargetVectorJoinEnum.MANUAL_WEIGHTS,
            target_vectors=list(weights.keys()),
            weights=weights,
        )

    @staticmethod
    def relative_score(weights: Dict[str, float]) -> _MultiTargetVectorJoin:
        """Combine the distance from different target vectors using score fusion."""
        return _MultiTargetVectorJoin(
            combination=_MultiTargetVectorJoinEnum.RELATIVE_SCORE,
            target_vectors=list(weights.keys()),
            weights=weights,
        )


class HybridVector:
    """Use this factory class to define the appropriate classes needed when defining near text and near vector sub-searches in hybrid queries."""

    @staticmethod
    def near_text(
        query: Union[str, List[str]],
        *,
        certainty: Optional[float] = None,
        distance: Optional[float] = None,
        move_to: Optional[Move] = None,
        move_away: Optional[Move] = None,
    ) -> _HybridNearText:
        """Define a near text search to be used within a hybrid query.

        Arguments:
            `query`
                The text to search for as a string or a list of strings.
            `certainty`
                The minimum similarity score to return. If not specified, the default certainty specified by the server is used.
            `distance`
                The maximum distance to search. If not specified, the default distance specified by the server is used.
            `move_to`
                Define the concepts that should be moved towards in the vector space during the search.
            `move_away`
                Define the concepts that should be moved away from in the vector space during the search.

        Returns:
            A `_HybridNearText` object to be used in the `vector` parameter of the `query.hybrid` and `generate.hybrid` search methods.
        """
        return _HybridNearText(
            text=query,
            distance=distance,
            certainty=certainty,
            move_to=move_to,
            move_away=move_away,
        )

    @staticmethod
    def near_vector(
        vector: List[float],
        *,
        certainty: Optional[float] = None,
        distance: Optional[float] = None,
    ) -> _HybridNearVector:
        """Define a near vector search to be used within a hybrid query.

        Arguments:
            `certainty`
                The minimum similarity score to return. If not specified, the default certainty specified by the server is used.
            `distance`
                The maximum distance to search. If not specified, the default distance specified by the server is used.

        Returns:
            A `_HybridNearVector` object to be used in the `vector` parameter of the `query.hybrid` and `generate.hybrid` search methods.
        """
        return _HybridNearVector(vector=vector, distance=distance, certainty=certainty)


NearVectorInputType = Union[List[float], Dict[str, List[float]], List[List[float]]]


class _QueryReference(_WeaviateInput):
    link_on: str
    include_vector: bool = Field(default=False)
    return_metadata: Optional[MetadataQuery] = Field(default=None)
    return_properties: Optional["PROPERTIES"] = Field(default=None)
    return_references: Optional["REFERENCES"] = Field(default=None)

    def __hash__(self) -> int:  # for set
        return hash(str(self))

    @property
    def _return_metadata(self) -> _MetadataQuery:
        return _MetadataQuery.from_public(self.return_metadata, self.include_vector)


class _QueryReferenceMultiTarget(_QueryReference):
    target_collection: str


class QueryReference(_QueryReference):
    """Define a query-time reference to a single-target property when querying through cross-references."""

    MultiTarget: ClassVar[Type[_QueryReferenceMultiTarget]] = _QueryReferenceMultiTarget
    """Define a query-time reference to a multi-target property when querying through cross-references."""


class QueryNested(_WeaviateInput):
    """Define the query-time return properties of a nested property."""

    name: str
    properties: "PROPERTIES"

    def __hash__(self) -> int:  # for set
        return hash(str(self))


REFERENCE = Union[_QueryReference, _QueryReferenceMultiTarget]
REFERENCES = Union[Sequence[REFERENCE], REFERENCE]

PROPERTY = Union[str, QueryNested]
PROPERTIES = Union[Sequence[PROPERTY], PROPERTY]

NestedProperties = Union[List[Union[str, QueryNested]], str, QueryNested]


class NearMediaType(str, Enum):
    """The different types of media that can be used in a `near_media` query to leverage the `multi2vec-*` modules.

    All are available when using `multi2vec-bind` but only `IMAGE` is available when using `multi2vec-clip`.
    """

    AUDIO = "audio"
    DEPTH = "depth"
    IMAGE = "image"
    IMU = "imu"
    THERMAL = "thermal"
    VIDEO = "video"



__all__ = ["_BatchClient", "_BatchCollection", "_BatchGRPC", "_BatchREST"]

from .client import _BatchClient
from .collection import _BatchCollection
from .grpc_batch_objects import _BatchGRPC
from .rest import _BatchREST



from .generate import _NearVectorGenerateAsync, _NearVectorGenerate
from .query import _NearVectorQueryAsync, _NearVectorQuery

__all__ = [
    "_NearVectorGenerate",
    "_NearVectorQuery",
    "_NearVectorGenerateAsync",
    "_NearVectorQueryAsync",
]



from typing import List, Literal, Optional, overload

from weaviate.collections.classes.cluster import Node, Shards, Stats
from weaviate.collections.cluster.cluster import _ClusterBase

class _Cluster(_ClusterBase):
    @overload
    def nodes(
        self,
        collection: Optional[str] = None,
        *,
        output: Literal[None] = None,
    ) -> List[Node[None, None]]: ...
    @overload
    def nodes(
        self,
        collection: Optional[str] = None,
        *,
        output: Literal["minimal"],
    ) -> List[Node[None, None]]: ...
    @overload
    def nodes(
        self,
        collection: Optional[str] = None,
        *,
        output: Literal["verbose"],
    ) -> List[Node[Shards, Stats]]: ...
